"""
This file contains classes and methods for option pricing and analysis, utilizing various financial models and numerical methods. 
It allows for calculations of option Greeks and prices either using tree methods or finite difference methods among others. 
It supports both European and American options for price calculation.
Classes:
- Greeks: Calculate option Greeks (Delta, Gamma, Vega, etc.) using the Black-Scholes formula.
- OptionPricing: Main class for option pricing that utilizes different numerical methods and models including GARCH for volatility modeling.
The design favors flexibility in choosing pricing methods and models, offering tools for both simple and advanced option pricing scenarios.
"""
from stockdatamanager.customerrors import MethodError
from .datafetcher import Fetcher
from .utils import _SMA, _EMA, _SARIMAX_model
import pandas as pd
import numpy as np
import yfinance as yf
import re
from typing import Union
from scipy.stats import norm
from scipy import linalg
from scipy.stats import linregress
import matplotlib.pyplot as plt
import seaborn as sns
import arch
import optuna
import math
from abc import ABC, abstractmethod
import sys

class Greeks(object):
  """
  Class to calculate the greeks for an option
  Inputs:
  - ticker: str, the ticker of the stock
  - call: bool, whether the option is a call or a put
  - identification: Union[str, int], the contract symbol of the option or the index of the option in the option chain
                                     Note: to find the contract identifier, it is written as ticker-YY-MM-DD-C(or P)-strike price (omit the - and the strike price is in a thousands format, so 1000 is $1.00)
  - verbose: int, the verbosity of the output
  """
  def __init__(self, 
               ticker: str,
               call: bool,
               identification: Union[str, int],
               verbose: bool = False):
    self.fetcher = Fetcher(ticker)
    self.df, self.yf_ticker = self.fetcher.df, self.fetcher.yf_stock
    if type(identification) == str:
     self.option_chain = self.yf_ticker.option_chain().calls if call else self.yf_ticker.option_chain().puts
     self.option = self.option_chain[self.option_chain['contractSymbol'] == identification]
    elif type(identification) == int:
      self.option = self.yf_ticker.option_chain().calls.iloc[identification] if call else self.yf_ticker.option_chain().puts.iloc[identification]
    else:
      raise ValueError('identification must be either a string or an integer')
    self.verbose = verbose
    self.date = self.calculate_expiration()
    self.call = call

  def from_name_to_datestr(self, s: str) -> str:
    """Helper function to convert the contract symbol to a date string"""
    if self.verbose:
      print(f"Processing {s}")
    match = re.search(r'[A-Za-z](\d{2})(\d{2})(\d{2})[CP]', s)
    if match:
        year, month, day = match.groups()
        return f"20{year}-{month}-{day}"
    else:
        return "No date found"
  
  def calculate_expiration(self) -> pd.Timestamp:
    """Function to calculate the expiration date of the option"""
    if self.verbose:
      print("Calculating expiration date")
    to_process = self.option['contractSymbol']
    date_str = self.from_name_to_datestr(to_process)
    date = pd.to_datetime(date_str)
    return date

  def calculate_delta(self, risk_free_rate_horizon: str = '13 weeks') -> float:
    """
    Method to calculate the delta of the option fed into the class, using the Black-Scholes formula.
    The delta is calculated as the first derivative of the option price with respect to the price of the underlying asset.
    Inputs:
    - risk_free_rate_horizon: str, the horizon of the risk free rate
    """
    if self.verbose:
      print("Calculating delta")
    option = self.option
    dataframe = self.df
    if self.call:
      S = dataframe['Close'].iloc[-1]
      K = option['strike']
      ##convert self.date to days until expiration
      T = (self.date - pd.to_datetime('today')).days / 365
      r = self.fetcher.get_risk_free_rate(risk_free_rate_horizon)['Close'].iloc[-1]
      sigma = option['impliedVolatility']
      d1 = (np.log(S/K) + (r + sigma**2 / 2) * T) / (sigma * np.sqrt(T))
      delta = norm.cdf(d1)
      return delta
    else:
      S = dataframe['Close'].iloc[-1]
      K = option['strike']
      T = (self.date - pd.to_datetime('today')).days / 365
      r = self.fetcher.get_risk_free_rate(risk_free_rate_horizon)['Close'].iloc[-1]
      sigma = option['impliedVolatility']
      d1 = (np.log(S / K) + (r + sigma ** 2 / 2) * T) / (sigma * np.sqrt(T))
      delta = norm.cdf(d1) - 1
      return delta
    
  def calculate_gamma(self, risk_free_rate_horizon: str = '13 weeks') -> float:
    """
    Method to calculate the gamma of the option fed into the class, using the Black-Scholes formula.
    The gamma is calculated as the second derivative of the option price with respect to the price of the underlying asset.
    Inputs:
    - risk_free_rate_horizon: str, the horizon of the risk free rate
    """
    if self.verbose:
      print("Calculating gamma")
    option = self.option
    dataframe = self.df
    S = dataframe['Close'].iloc[-1]
    K = option['strike']
    T = (self.date - pd.to_datetime('today')).days / 365
    r = self.fetcher.get_risk_free_rate(risk_free_rate_horizon)['Close'].iloc[-1]
    sigma = option['impliedVolatility']
    d1 = (np.log(S / K) + (r + sigma ** 2 / 2) * T) / (sigma * np.sqrt(T))
    gamma = norm.pdf(d1) / (S * sigma * np.sqrt(T))
    return gamma
  
  def calculate_vega(self, risk_free_rate_horizon: str = '13 weeks') -> float:
    """
    Method to calculate the vega of the option fed into the class, using the Black-Scholes formula.
    The vega is calculated as the first derivative of the option price with respect to the implied volatility.
    Inputs:
    - risk_free_rate_horizon: str, the horizon of the risk free rate
    """
    if self.verbose:
      print("Calculating vega")
    option = self.option
    dataframe = self.df
    S = dataframe['Close'].iloc[-1]
    K = option['strike']
    T = (self.date - pd.to_datetime('today')).days / 365
    r = self.fetcher.get_risk_free_rate(risk_free_rate_horizon)['Close'].iloc[-1]
    sigma = option['impliedVolatility']
    d1 = (np.log(S / K) + (r + sigma ** 2 / 2) * T) / (sigma * np.sqrt(T))
    vega = S * norm.pdf(d1) * np.sqrt(T)
    return vega
  
  def calculate_vanna(self, risk_free_rate_horizon: str = '13 weeks') -> float:
    """
    Method to calculate the vanna of the option fed into the class, using the Black-Scholes formula.
    The vanna is calculated as the first derivative of the delta with respect to the implied volatility.
    Inputs:
    - risk_free_rate_horizon: str, the horizon of the risk free rate
    """
    if self.verbose:
      print("Calculating vanna")
    option = self.option
    dataframe = self.df
    S = dataframe['Close'].iloc[-1]
    K = option['strike']
    T = (self.date - pd.to_datetime('today')).days / 365
    r = self.fetcher.get_risk_free_rate(risk_free_rate_horizon)['Close'].iloc[-1]
    sigma = option['impliedVolatility']
    d1 = (np.log(S / K) + (r + sigma ** 2 / 2) * T) / (sigma * np.sqrt(T))
    vanna = -norm.pdf(d1) * d1 / sigma
    return vanna
  
  def calculate_theta(self, risk_free_rate_horizon: str = '13 weeks') -> float:
    """
    Method to calculate the theta of the option fed into the class, using the Black-Scholes formula.
    The theta is calculated as the first derivative of the option price with respect to time.
    Inputs:
    - risk_free_rate_horizon: str, the horizon of the risk free rate
    """
    if self.verbose:
      print("Calculating theta")
    option = self.option
    dataframe = self.df
    S = dataframe['Close'].iloc[-1]
    K = option['strike']
    T = (self.date - pd.to_datetime('today')).days / 365
    r = self.fetcher.get_risk_free_rate(risk_free_rate_horizon)['Close'].iloc[-1]
    sigma = option['impliedVolatility']
    dividends = self.fetcher.get_dividend_yield()
    r_adj = r - dividends
    d1 = (np.log(S / K) + (r_adj + 0.5 * sigma ** 2) * T) / (sigma * np.sqrt(T))
    d2 = d1 - sigma * np.sqrt(T)
    
    if self.call:
        theta = - (S * sigma * norm.pdf(d1)) / (2 * np.sqrt(T)) - r * K * np.exp(-r * T) * norm.cdf(d2)
    else:
        theta = - (S * sigma * norm.pdf(d1)) / (2 * np.sqrt(T)) + r * K * np.exp(-r * T) * norm.cdf(-d2)    
    theta_per_day = theta / 365
    return theta_per_day
  
  def calculate_rho(self, risk_free_rate_horizon: str = '13 weeks') -> float:
    """
    Method to calculate the rho of the option fed into the class, using the Black-Scholes formula.
    The rho is calculated as the first derivative of the option price with respect to the risk-free rate.
    Inputs:
    - risk_free_rate_horizon: str, the horizon of the risk free rate
    """
    if self.verbose:
      print("Calculating rho")
    option = self.option
    dataframe = self.df
    S = dataframe['Close'].iloc[-1]
    K = option['strike']
    T = (self.date - pd.to_datetime('today')).days / 365
    r = self.fetcher.get_risk_free_rate(risk_free_rate_horizon)['Close'].iloc[-1]
    sigma = option['impliedVolatility']
    d1 = (np.log(S / K) + (r + sigma ** 2 / 2) * T) / (sigma * np.sqrt(T))
    d2 = d1 - sigma * np.sqrt(T)
    if self.call:
      rho = K * T * np.exp(-r * T) * norm.cdf(d2)
    else:
      rho = -K * T * np.exp(-r * T) * norm.cdf(-d2)
    return rho
  
  def calculate_vomma(self, risk_free_rate_horizon: str = '13 weeks') -> float:
    """
    Method to calculate the vomma of the option fed into the class, using the Black-Scholes formula.
    The vomma is calculated as the second derivative of the option price with respect to the implied volatility.
    Inputs:
    - risk_free_rate_horizon: str, the horizon of the risk free rate
    """
    if self.verbose:
      print("Calculating vomma")
    option = self.option
    dataframe = self.df
    S = dataframe['Close'].iloc[-1]
    K = option['strike']
    T = (self.date - pd.to_datetime('today')).days / 365
    r = self.fetcher.get_risk_free_rate(risk_free_rate_horizon)['Close'].iloc[-1]
    sigma = option['impliedVolatility']
    d1 = (np.log(S / K) + (r + 0.5 * sigma ** 2) * T) / (sigma * np.sqrt(T))
    d2 = d1 - sigma * np.sqrt(T)
    vomma = S * np.sqrt(T) * norm.pdf(d1) * (d1 * d2) / sigma
    return vomma
  
  def calculate_charm(self, risk_free_rate_horizon: str = '13 weeks') -> float:
    """
    Method to calculate the charm of the option fed into the class, using the Black-Scholes formula.
    The charm is calculated as the first derivative of the delta with respect to time.
    Inputs:
    - risk_free_rate_horizon: str, the horizon of the risk free rate
    """
    if self.verbose:
      print("Calculating charm")
    option = self.option
    dataframe = self.df
    S = dataframe['Close'].iloc[-1]
    K = option['strike']
    T = (self.date - pd.to_datetime('today')).days / 365
    r = self.fetcher.get_risk_free_rate(risk_free_rate_horizon)['Close'].iloc[-1]
    sigma = option['impliedVolatility']
    d1 = (np.log(S / K) + (r + sigma ** 2 / 2) * T) / (sigma * np.sqrt(T))
    d2 = d1 - sigma * np.sqrt(T)
    charm = -norm.pdf(d1) * (2 * r * T - d2 * sigma * np.sqrt(T)) / (2 * T * sigma * np.sqrt(T))
    return charm
  
  def calculate_speed(self, risk_free_rate_horizon: str = '13 weeks') -> float:
    """
    Method to calculate the speed of the option fed into the class, using the Black-Scholes formula.
    The speed is calculated as the third derivative of the option price with respect to the price of the underlying asset.
    Inputs:
    - risk_free_rate_horizon: str, the horizon of the risk free rate
    """
    if self.verbose:
      print("Calculating speed")
    option = self.option
    dataframe = self.df
    S = dataframe['Close'].iloc[-1]
    K = option['strike']
    T = (self.date - pd.to_datetime('today')).days / 365
    r = self.fetcher.get_risk_free_rate(risk_free_rate_horizon)['Close'].iloc[-1]
    sigma = option['impliedVolatility']
    d1 = (np.log(S / K) + (r + sigma ** 2 / 2) * T) / (sigma * np.sqrt(T))
    speed = -norm.pdf(d1) / (S ** 2 * sigma * np.sqrt(T)) * (d1 + sigma * np.sqrt(T))
    return speed
  
  def calculate_zomma(self, risk_free_rate_horizon: str = '13 weeks') -> float:
    """
    Method to calculate the zomma of the option fed into the class, using the Black-Scholes formula.
    The zomma is calculated as the derivative of the gamma with respect to the price of the underlying asset.
    Inputs:
    - risk_free_rate_horizon: str, the horizon of the risk free rate
    """
    if self.verbose:
      print("Calculating zomma")
    option = self.option
    dataframe = self.df
    S = dataframe['Close'].iloc[-1]
    K = option['strike']
    T = (self.date - pd.to_datetime('today')).days / 365
    r = self.fetcher.get_risk_free_rate(risk_free_rate_horizon)['Close'].iloc[-1]
    sigma = option['impliedVolatility']
    d1 = (np.log(S / K) + (r + sigma ** 2 / 2) * T) / (sigma * np.sqrt(T))
    d2 = d1 - sigma * np.sqrt(T)
    zomma = norm.pdf(d1) * (d1 * d2 - 1) / (S * sigma ** 2 * np.sqrt(T))
    return zomma
  
  def calculate_color(self, risk_free_rate_horizon: 'str' = '13 weeks') -> float:
    """
    Method to calculate the color of the option fed into the class, using the Black-Scholes formula.
    The color is calculated as the second derivative of the charm with respect to the price of the underlying asset.
    Inputs:
    - risk_free_rate_horizon: str, the horizon of the risk free rate
    """
    if self.verbose:
      print("Calculating color")
    option = self.option
    dataframe = self.df
    S = dataframe['Close'].iloc[-1]
    K = option['strike']
    T = (self.date - pd.to_datetime('today')).days / 365
    r = self.fetcher.get_risk_free_rate(risk_free_rate_horizon)['Close'].iloc[-1]
    sigma = option['impliedVolatility']
    d1 = (np.log(S/K) + (r + sigma**2 / 2) * T) / (sigma * np.sqrt(T))
    d2 = d1 - sigma * np.sqrt(T)
    color = -norm.pdf(d1) * (2 * r * T - d2 * sigma * np.sqrt(T)) / (2 * T * sigma**2 * np.sqrt(T))
    return color

  def calculate_ultima(self, risk_free_rate_horizon: 'str' = '13 weeks') -> float:
    """
    Method to calculate the ultima of the option fed into the class, using the Black-Scholes formula.
    The ultima is calculated as the second derivative of the vomma with respect to the price of the underlying asset.
    Inputs:
    - risk_free_rate_horizon: str, the horizon of the risk free rate
    """
    if self.verbose:
      print("Calculating ultima")
    option = self.option
    dataframe = self.df
    S = dataframe['Close'].iloc[-1]
    K = option['strike']
    T = (self.date - pd.to_datetime('today')).days / 365
    r = self.fetcher.get_risk_free_rate(risk_free_rate_horizon)['Close'].iloc[-1]
    sigma = option['impliedVolatility']
    d1 = (np.log(S/K) + (r + sigma**2 / 2) * T) / (sigma * np.sqrt(T))
    d2 = d1 - sigma * np.sqrt(T)
    ultima = S * np.sqrt(T) * norm.pdf(d1) * (d1 * d2 * (1 - d1 * d2) + d1 ** 2 + d2 ** 2) / sigma ** 2
    return ultima
  
  def calculate_lambda(self,risk_free_rate_horizon: str = '13 weeks') -> float:
    """
    Method to calculate the lambda (leverage) of the option fed into the class.
    Lambda is calculated as the product of Delta and the ratio of the underlying asset's price to the option's price.
    """
    if self.verbose:
        print("Calculating lambda")
    option = self.option
    dataframe = self.df
    S = dataframe['Close'].iloc[-1]  
    V = option['lastPrice'] 
    delta = self.calculate_delta(risk_free_rate_horizon)  
    lambda_ = delta * (S / V)
    return lambda_

class Option(object):
  def __init__(self, S, K, r, T, N, pu, pd_, div, sigma, call=True, american=False, verbose = False):
    """
    Inputs:
    - S: float, the price of the underlying asset
    - K: float, the strike price of the option
    - r: float, the risk-free rate
    - T: float, the time to expiration of the option
    - N: int, the number of time steps
    - pu: float, the probability at the up state
    - pd_: float, the probability at the down state
    - div: float, the dividend yield
    - sigma: float, the volatility of the underlying asset
    - call: bool, whether the option is a call or a put
    - american: bool, whether the option is American or European
    - verbose: bool, the verbosity of the output
    Internal class to implement all the tree methods for option pricing
    Please do not use this class, it is for internal use only (I know, I should have put a _ in the name, but I didn't, so please don't use it)
    """
    self.S = S
    self.K = K
    self.r = r
    self.T = T
    self.N = max(1, N)
    self.STs = [] 
    self.pu, self.pd_= pu, pd_
    self.div = div
    self.sigma = sigma
    self.is_call = call
    self.is_european = not american
    self.verbose = verbose

  @property
  def dt(self) -> float:
    """ Single time step, in years """
    return self.T / float(self.N)

  @property
  def df(self) -> float:
    """ The discount factor """
    return math.exp(-(self.r - self.div) * self.dt)  

class _FiniteDifferences(object):
  def __init__(self, S, K, r, T, sigma, Smax, M, N, call, verbose = False):
    """
    Inputs:
    - S: float, the price of the underlying asset
    - K: float, the strike price of the option
    - r: float, the risk-free rate
    - T: float, the time to expiration of the option
    - sigma: float, the volatility of the underlying asset
    - Smax: float, the maximum price of the underlying asset
    - M: int, the number of price steps
    - N: int, the number of time steps
    - call: bool, whether the option is a call or a put
    - verbose: bool, the verbosity of the output
    Internal class to implement the Finite Difference Method for option pricing
    Please do not use this class, it is for internal use only (I know, I should have put a _ in the name, but I didn't, so please don't use it)
    """
    self.S = S
    self.K = K
    self.r = r
    self.T = T
    self.sigma = sigma
    self.Smax = Smax
    self.M, self.N = M, N
    self.is_call = call
    self.i_values = np.arange(self.M)
    self.j_values = np.arange(self.N)
    self.grid = np.zeros(shape=(self.M + 1, self.N + 1))
    self.boundary_conds = np.linspace(0, Smax, self.M+1)
    self.verbose = verbose

  @property
  def dS(self) -> float:
    """  Single step in S space"""
    return self.Smax / float(self.M)

  @property
  def dt(self) -> float:
    """" Single step in t space"""
    return self.T / float(self.N)

  @abstractmethod
  def setup_boundary_conditions(self):
    raise NotImplementedError('This method is implemented in the subclasses')

  @abstractmethod
  def setup_coefficients(self):
    raise NotImplementedError('This method is implemented in the subclasses')

  @abstractmethod
  def traverse_grid(self):
    raise NotImplementedError("This method is implemented in the subclasses")
  @abstractmethod

  def interpolate(self) -> float:
    """
    Use piecewise linear interpolation on the initial grid column to get the closest price at S.
    """
    if self.verbose: print("Interpolating the option price")
    return np.interp(self.S, self.boundary_conds, self.grid[:,0])

  def price(self) -> float:
    """ Entry point of the pricing method"""
    if self.verbose: print("Pricing the option")
    self.setup_boundary_conditions()
    self.setup_coefficients()
    self.traverse_grid()
    return self.interpolate()

class _FDEU(_FiniteDifferences):
  """
  Class to implement the Explicit Finite Difference Method for option pricing,
  Meant for internal use only
  """
  def setup_boundary_conditions(self):
    """Setup the boundary conditions for the option pricing problem"""
    if self.verbose: print("Setting up boundary conditions")
    if self.is_call:
      self.grid[:,-1] = np.maximum(0, self.boundary_conds - self.K)
      self.grid[-1,:-1] = (self.Smax-self.K) * np.exp(-self.r*self.dt*(self.N - self.j_values))
    else:
      self.grid[:,-1] = np.maximum(0, self.K-self.boundary_conds)
      self.grid[0,:-1] = (self.K-self.Smax) * np.exp(-self.r*self.dt*(self.N-self.j_values))
    if self.verbose: print("Boundary conditions set")

  def setup_coefficients(self):
    """Setup the coefficients for the finite difference scheme"""
    if self.verbose: print("Setting up coefficients")
    self.a = 0.5 * self.dt * ((self.sigma ** 2) * (self.i_values ** 2) - self.r * self.i_values)
    self.b = 1 - self.dt * ((self.sigma ** 2) *(self.i_values ** 2) + self.r)
    self.c = 0.5 * self.dt * ((self.sigma ** 2) * (self.i_values ** 2) + self.r * self.i_values)
    if self.verbose: print("Coefficients set")

  def traverse_grid(self):
    """Function to iterate the grid backwards in time"""
    if self.verbose: print("Traversing the grid")
    for j in reversed(self.j_values):
      for i in range(self.M)[2:]:
        self.grid[i,j] = self.a[i]*self.grid[i-1,j+1] + self.b[i]*self.grid[i,j+1] + self.c[i]*self.grid[i+1,j+1]
    if self.verbose: print("Grid traversed")
                  
class _FDImpEU(_FDEU):
  """
  Internal class to implement the Implicit Finite Difference Method for option pricing
  """
  def setup_coefficients(self):
    """Setup the coefficients for the implicit finite difference scheme"""
    if self.verbose: print("Setting up coefficients")
    self.a = 0.5 * (self.r * self.dt * self.i_values - (self.sigma ** 2) * self.dt * (self.i_values ** 2))
    self.b = 1 + (self.sigma ** 2) * self.dt * (self.i_values ** 2) + self.r * self.dt
    self.c = -0.5 * (self.r * self.dt * self.i_values + (self.sigma ** 2) * self.dt * (self.i_values ** 2))
    self.coeffs = np.diag(self.a[2:self.M],-1) + np.diag(self.b[1:self.M]) + np.diag(self.c[1:self.M-1],1)
    if self.verbose: print("Coefficients set")

  def traverse_grid(self):
    """ Solve using linear systems of equations """
    if self.verbose: print("Traversing the grid")
    P, L, U = linalg.lu(self.coeffs)
    aux = np.zeros(self.M-1)
    for i in reversed(range(self.N)):
      aux[0] = np.dot(-self.a[1], self.grid[0, i])
      x1 = linalg.solve(L, self.grid[1:self.M, i+1]+aux)
      x2 = linalg.solve(U, x1)
      self.grid[1:self.M, i] = x2
    if self.verbose: print("Grid traversed")

class _FDCN(_FDEU):
  """Internal class to implement the Crank-Nicolson Finite Difference Method for option pricing"""
  def setup_coefficients(self):
    """Setup the coefficients for the Crank-Nicolson scheme"""
    if self.verbose: print("Setting up coefficients")
    self.alpha = 0.25 * self.dt * ((self.sigma ** 2) * (self.i_values ** 2) - self.r * self.i_values)
    self.beta = -self.dt * 0.5 * ((self.sigma ** 2) * (self.i_values ** 2) + self.r)
    self.gamma = 0.25 * self.dt * ((self.sigma ** 2) * (self.i_values ** 2) + self.r * self.i_values)
    self.M1 = -np.diag(self.alpha[2:self.M], -1) + np.diag(1-self.beta[1:self.M]) - np.diag(self.gamma[1:self.M-1], 1)
    self.M2 = np.diag(self.alpha[2:self.M], -1) + np.diag(1+self.beta[1:self.M]) + np.diag(self.gamma[1:self.M-1], 1)
    if self.verbose: print("Coefficients set")
  
  def traverse_grid(self):
    """ Solve using linear systems of equations """
    if self.verbose: print("Traversing the grid")
    P, L, U = linalg.lu(self.M1)
    for j in reversed(range(self.N)):
      x1 = linalg.solve(L, np.dot(self.M2, self.grid[1:self.M, j+1]))
      x2 = linalg.solve(U, x1)
      self.grid[1:self.M, j] = x2
    if self.verbose: print("Grid traversed")

class _FDCNUS(_FDCN):
  """Internal class to implement the Crank-Nicolson Finite Difference Method for american option pricing"""
  def __init__(self, S, K, r, T, sigma, Smax, M, N, omega, tol, call, verbose = False):
    super(_FDCNUS, self).__init__(S, K, r=r, T=T, sigma=sigma, Smax=Smax, M=M, N=N, call=call, verbose=verbose)
    self.omega = omega
    self.tol = tol
    self.i_values = np.arange(self.M+1)
    self.j_values = np.arange(self.N+1)

  def setup_boundary_conditions(self):
    """Setup the boundary conditions for the option pricing problem"""
    if self.verbose: print("Setting up boundary conditions")
    if self.is_call:
      self.payoffs = np.maximum(0, self.boundary_conds[1:self.M] - self.K)
    else:
      self.payoffs = np.maximum(0, self.K - self.boundary_conds[1:self.M])
    self.past_values = self.payoffs
    self.boundary_values = self.K * np.exp(-self.r * self.dt * (self.N - self.j_values))
    if self.verbose: print("Boundary conditions set")
        
  def calculate_payoff_start_boundary(self, rhs, old_values):
    """Calculate the payoff at the starting boundary"""
    payoff = old_values[0] + self.omega / (1 - self.beta[1]) * (rhs[0] - (1 - self.beta[1]) * old_values[0] + self.gamma[1] * old_values[1])
    return max(self.payoffs[0], payoff)      
    
  def calculate_payoff_end_boundary(self, rhs, old_values, new_values):
    """Calculate the payoff at the ending boundary"""
    payoff = old_values[-1] + self.omega / (1 - self.beta[-2]) * (rhs[-1] + self.alpha[-2] * new_values[-2] - (1 - self.beta[-2]) * old_values[-1])
    return max(self.payoffs[-1], payoff)
    
  def calculate_payoff(self, k, rhs, old_values, new_values):
    """Calculate the payoff at the k-th point"""
    payoff = old_values[k] + self.omega / (1 - self.beta[k+1]) * (rhs[k] + self.alpha[k + 1] * new_values[k - 1] - (1 - self.beta[k + 1]) * old_values[k] + self.gamma[k + 1] * old_values[k + 1])
    return max(self.payoffs[k], payoff)

  def traverse_grid(self):
    """ Solve using linear systems of equations """
    if self.verbose: print("Traversing the grid")
    aux = np.zeros(self.M - 1)
    new_values = np.zeros(self.M - 1)
    for j in reversed(range(self.N)):
      aux[0] = self.alpha[1] * (self.boundary_values[j] + self.boundary_values[j + 1])
      rhs = np.dot(self.M2, self.past_values) + aux
      old_values = np.copy(self.past_values)
      error = sys.float_info.max
      while self.tol < error:
        new_values[0] = self.calculate_payoff_start_boundary(rhs, old_values)            
        for k in range(self.M-2)[1:]:
          new_values[k] = self.calculate_payoff(k, rhs, old_values, new_values)                    
        new_values[-1] = self.calculate_payoff_end_boundary(rhs, old_values, new_values)
        error = np.linalg.norm(new_values-old_values)
        old_values = np.copy(new_values)
        self.past_values = np.copy(new_values)
    self.values = np.concatenate(
        ([self.boundary_values[0]], new_values, [0]))
    if self.verbose: print("Grid traversed")

  def interpolate(self):
    """ Use piecewise linear interpolation on the initial grid column to get the closest price at S."""
    if self.verbose: print("Interpolating the option price")
    return np.interp(self.S, self.boundary_conds, self.values)

class _BinomialCRRLattice(Option):
  def setup_parameters(self):
    """ Required calculations for the model"""
    if self.verbose: print("Setting up parameters")
    self.u = math.exp(self.sigma * math.sqrt(self.dt))
    self.d = 1 / self.u
    self.qu = (math.exp((self.r - self.div)*self.dt) - self.d) / (self.u-self.d)
    self.qd = 1 - self.qu
    self.M = 2 * self.N + 1
    if self.verbose: print("Parameters set up")

  def init_stock_price_tree(self):
    """ Function for the initialization of the stock price tree"""
    if self.verbose: print("Initializing stock price tree")
    self.STs = np.zeros(self.M)
    self.STs[0] = self.S * self.u ** self.N
    for i in range(self.M)[1:]:
      self.STs[i] = self.STs[i-1]*self.d
    if self.verbose: print("Stock price tree initialized")
  
  def init_payoffs_tree(self):
    """ Function for the initialization of the payoffs tree"""
    if self.verbose: print("Initializing payoffs tree")
    odd_nodes = self.STs[::2]  
    if self.is_call:
        return np.maximum(0, odd_nodes-self.K)
    else:
        return np.maximum(0, self.K-odd_nodes)
    
  def check_early_exercise(self, payoffs):
    """ Function to check for early exercise of the option"""
    self.STs = self.STs[1:-1] 
    odd_STs = self.STs[::2]
    if self.is_call:
        return np.maximum(payoffs, odd_STs-self.K)
    else:
        return np.maximum(payoffs, self.K-odd_STs)
    
  def traverse_tree(self, payoffs):
    """ Function to traverse the tree and calculate the option price"""
    if self.verbose: print("Traversing tree")
    for i in reversed(range(self.N)):
      payoffs = (payoffs[:-1]*self.qu + 
                 payoffs[1:]*self.qd)*self.df
      if not self.is_european:
          payoffs = self.check_early_exercise(payoffs)
    if self.verbose: print("Tree traversal complete")
    return payoffs
  
  def begin_tree_traversal(self):
    if self.verbose: print("Beginning tree traversal")
    payoffs = self.init_payoffs_tree()
    return self.traverse_tree(payoffs)
  
  def price(self):
    if self.verbose: print("Pricing the option")
    self.setup_parameters()
    self.init_stock_price_tree()
    payoffs = self.begin_tree_traversal()
    if self.verbose: print("Option priced")
    return payoffs[0]

class _TrinomialLattice(Option):
  def setup_parameters(self):
    """ Required calculations for the model """
    if self.verbose: print("Setting up parameters")
    self.u = math.exp(self.sigma * math.sqrt(2 * self.dt))
    self.d = 1/self.u
    self.m = 1
    self.qu = ((math.exp((self.r-self.div) * self.dt/2) - math.exp(-self.sigma * math.sqrt(self.dt / 2))) / (math.exp(self.sigma * math.sqrt(self.dt / 2)) - math.exp(-self.sigma * math.sqrt(self.dt / 2)))) ** 2
    self.qd = ((math.exp(self.sigma * math.sqrt(self.dt/2)) - math.exp((self.r - self.div) * self.dt / 2)) / (math.exp(self.sigma * math.sqrt(self.dt / 2)) - math.exp(-self.sigma * math.sqrt(self.dt / 2)))) ** 2
    self.qm = 1 - self.qu - self.qd
    self.M = 2 * self.N + 1
    if self.verbose: print("Parameters set up")

  def init_stock_price_tree(self):
    """ Initialize the stock price tree """
    if self.verbose: print("Initializing stock price tree")
    self.STs = np.zeros(self.M)
    self.STs[0] = self.S * self.u**self.N
    for i in range(self.M)[1:]:
      self.STs[i] = self.STs[i-1]*self.d
    if self.verbose: print("Stock price tree initialized")

  def init_payoffs_tree(self):
    """ Initialize the payoffs tree """
    if self.verbose: print("Initializing payoffs tree")
    if self.is_call:
      return np.maximum(0, self.STs-self.K)
    else:
      return np.maximum(0, self.K-self.STs)
    
  def check_early_exercise(self, payoffs):
    """ Check for early exercise of the option """
    self.STs = self.STs[1:-1]  
    if self.is_call:
      return np.maximum(payoffs, self.STs-self.K)
    else:
      return np.maximum(payoffs, self.K-self.STs)
  
  def traverse_tree(self, payoffs):
    """ Traverse the tree and calculate the option price"""
    if self.verbose: print("Traversing tree")
    for i in reversed(range(self.N)):
      payoffs = (payoffs[:-2] * self.qu + payoffs[1:-1] * self.qm + payoffs[2:] * self.qd) * self.df
      if not self.is_european:
          payoffs = self.check_early_exercise(payoffs)
    if self.verbose: print("Tree traversal complete")
    return payoffs
  
  def begin_tree_traversal(self):
    """ Begin the tree traversal """
    if self.verbose: print("Beginning tree traversal")
    payoffs = self.init_payoffs_tree()
    return self.traverse_tree(payoffs)
  
  def price(self):
    """  The pricing implementation """
    if self.verbose: print("Pricing the option")
    self.setup_parameters()
    self.init_stock_price_tree()
    payoffs = self.begin_tree_traversal()
    if self.verbose: print("Option priced")
    return payoffs[0]

class _BinomialTree(Option):
  """Internal class to implement the Binomial Tree Method for option pricing"""
  def setup_parameters(self):
    if self.verbose: print("Setting up parameters")
    self.M = self.N + 1  
    self.u = 1 + self.pu
    self.d = 1 - self.pd_
    self.qu = (math.exp((self.r-self.div)*self.dt)-self.d)/(self.u-self.d)
    self.qd = 1-self.qu
    if self.verbose: print("Parameters set up")

  def init_stock_price_tree(self):
    """ Initialize the stock price tree """
    if self.verbose: print("Initializing stock price tree")
    self.STs = np.zeros(self.M)
    for i in range(self.M):
        self.STs[i] = self.S * (self.u**(self.N-i)) * (self.d**i)
    if self.verbose: print(f"Stock price tree initialized")

  def init_payoffs_tree(self):
    """
    Returns the payoffs when the option 
    expires at terminal nodes
    """
    if self.verbose: print("Initializing payoffs tree") 
    if self.is_call:
      return np.maximum(0, self.STs-self.K)
    else:
        return np.maximum(0, self.K-self.STs)
    
  def traverse_tree(self, payoffs):
    """
    Starting from the time the option expires, traverse
    backwards and calculate discounted payoffs at each node
    """
    if self.verbose: print("Traversing tree")
    for i in range(self.N):
      payoffs = (payoffs[:-1]*self.qu + payoffs[1:]*self.qd)*self.df
    if self.verbose: print("Tree traversal complete")
    return payoffs
  
  def begin_tree_traversal(self):
    """Traversal of the tree from the beginning"""
    if self.verbose: print("Beginning tree traversal")
    payoffs = self.init_payoffs_tree()
    return self.traverse_tree(payoffs)
  
  def price(self):
    """ Entry point of the pricing implementation """
    if self.verbose: print("Pricing the option")
    self.setup_parameters()
    self.init_stock_price_tree()
    payoffs = self.begin_tree_traversal()
    if self.verbose: print("Option priced")
    return payoffs[0]

class _USBinomialTree(_BinomialTree):
  def init_stock_price_tree(self):
    """ Initialize the stock price tree"""
    if self.verbose: print("Initializing stock price tree")
    self.STs = [np.array([self.S])]
    for i in range(self.N):
      prev_branches = self.STs[-1]
      st = np.concatenate(
          (prev_branches*self.u, 
           [prev_branches[-1]*self.d]))
      self.STs.append(st)
    if self.verbose: print("Stock price tree initialized")

  def init_payoffs_tree(self):
    """ Initialize the payoffs tree"""
    if self.verbose: print("Initializing payoffs tree")
    if self.is_call:
        return np.maximum(0, self.STs[self.N]-self.K)
    else:
        return np.maximum(0, self.K-self.STs[self.N])

  def check_early_exercise(self, payoffs, node):
    """ Check for early exercise of the option"""
    if self.is_call:
        return np.maximum(payoffs, self.STs[node] - self.K)
    else:
        return np.maximum(payoffs, self.K - self.STs[node])

  def traverse_tree(self, payoffs):
    """ Traverse the tree and calculate the option price"""
    if self.verbose: print("Traversing tree")
    for i in reversed(range(self.N)):
        payoffs = (payoffs[:-1]*self.qu + payoffs[1:]*self.qd)*self.df
        payoffs = self.check_early_exercise(payoffs,i)
    if self.verbose: print("Tree traversal complete")
    return payoffs
  
class _BinomialCRR(_USBinomialTree):
  def setup_parameters(self):
    """ Required calculations for the model """
    if self.verbose: print("Setting up parameters")
    self.u = math.exp(self.sigma * math.sqrt(self.dt))
    self.d = 1./self.u
    self.qu = (math.exp((self.r-self.div)*self.dt) - 
               self.d)/(self.u-self.d)
    self.qd = 1-self.qu
  
  def traverse_tree(self, payoffs):
    """ Traverse the tree and calculate the option price"""
    if self.verbose: print("Traversing tree")
    for i in reversed(range(self.N)):
        payoffs = (payoffs[:-1]*self.qu + payoffs[1:]*self.qd)*self.df
        if not self.is_european:
          payoffs = self.check_early_exercise(payoffs,i)
    if self.verbose: print("Tree traversal complete")
    return payoffs
  
class _BinomialLR(_USBinomialTree):
  def setup_parameters(self):
    """ Required calculations for the model """
    odd_N = self.N if (self.N%2 == 0) else (self.N+1)
    d1 = (math.log(self.S/self.K) +
          ((self.r-self.div) +
           (self.sigma**2)/2.)*self.T)/\
        (self.sigma*math.sqrt(self.T))
    d2 = (math.log(self.S/self.K) +
          ((self.r-self.div) -
           (self.sigma**2)/2.)*self.T)/\
        (self.sigma * math.sqrt(self.T))
    pbar = self.pp_2_inversion(d1, odd_N)
    self.p = self.pp_2_inversion(d2, odd_N)
    self.u = 1/self.df * pbar/self.p
    self.d = (1/self.df-self.p*self.u)/(1-self.p)
    self.qu = self.p
    self.qd = 1-self.p

  def pp_2_inversion(self, z, n):
    """ Peizer-Pratt inversion """
    if self.verbose: print("Calculating Peizer-Pratt inversion")
    return .5 + math.copysign(1, z)*\
        math.sqrt(.25 - .25*
            math.exp(
                -((z/(n+1./3.+.1/(n+1)))**2.)*(n+1./6.)
            )
        )  
  
  def traverse_tree(self, payoffs):
    """ Traverse the tree and calculate the option price"""
    if self.verbose: print("Traversing tree")
    for i in reversed(range(self.N)):
        payoffs = (payoffs[:-1]*self.qu + payoffs[1:]*self.qd)*self.df
        if not self.is_european:
          payoffs = self.check_early_exercise(payoffs,i)
    if self.verbose: print("Tree traversal complete")
    return payoffs 

class _TrinomialTree(_USBinomialTree):
  def setup_parameters(self):
    """ Required calculations for the model """
    if self.verbose: print("Setting up parameters")
    self.u = math.exp(self.sigma*math.sqrt(2.*self.dt))
    self.d = 1/self.u
    self.m = 1
    self.qu = ((math.exp((self.r-self.div) *
                         self.dt/2.) -
                math.exp(-self.sigma *
                         math.sqrt(self.dt/2.))) /
               (math.exp(self.sigma *
                         math.sqrt(self.dt/2.)) -
                math.exp(-self.sigma *
                         math.sqrt(self.dt/2.))))**2
    self.qd = ((math.exp(self.sigma *
                         math.sqrt(self.dt/2.)) -
                math.exp((self.r-self.div) *
                         self.dt/2.)) /
               (math.exp(self.sigma *
                         math.sqrt(self.dt/2.)) -
                math.exp(-self.sigma *
                         math.sqrt(self.dt/2.))))**2.
    self.qm = 1 - self.qu - self.qd
    if self.verbose: print("Parameters set up")

  def init_stock_price_tree(self):
    """ Initialize the stock price tree """
    if self.verbose: print("Initializing stock price tree")
    self.STs = [np.array([self.S])]
    for i in range(self.N):
      prev_nodes = self.STs[-1]
      self.ST = np.concatenate(
          (prev_nodes*self.u, [prev_nodes[-1]*self.m,
                               prev_nodes[-1]*self.d]))
      self.STs.append(self.ST)
    if self.verbose: print("Stock price tree initialized")

  def traverse_tree(self, payoffs):
    """ Traverse the tree and calculate the option price"""
    if self.verbose: print("Traversing tree")
    for i in reversed(range(self.N)):
      payoffs = (payoffs[:-2] * self.qu +
                 payoffs[1:-1] * self.qm +
                 payoffs[2:] * self.qd) * self.df
      if not self.is_european:
        payoffs = self.check_early_exercise(payoffs,i)
    if self.verbose: print("Tree traversal complete")
    return payoffs

class _BlackScholes(Option):
  def price(self):
    """ Calculate the option price using the Black-Scholes formula """
    if self.verbose: print("Calculating the option price using the Black-Scholes formula")
    d1 = (math.log(self.S/self.K) + ((self.r-self.div) + (self.sigma**2)/2.)*self.T) / (self.sigma*math.sqrt(self.T))
    d2 = d1 - self.sigma * math.sqrt(self.T)
    if self.is_call:
      return self.S * math.exp(-self.div*self.T) * norm.cdf(d1) - self.K * math.exp(-self.r*self.T) * norm.cdf(d2)
    else:
      return self.K * math.exp(-self.r*self.T) * norm.cdf(-d2) - self.S * math.exp(-self.div*self.T) * norm.cdf(-d1)

class _MonteCarlo(Option):
  """Internal class to implement the Monte Carlo Method for option pricing"""
  def __init__(self, S, K, r, T, N, pu, pd_, div, sigma, num_of_paths, call=True, american=False, verbose = False, display = False):
    """
    Inputs:
    - S: float, the price of the underlying asset
    - K: float, the strike price of the option
    - r: float, the risk-free rate
    - T: float, the time to expiration of the option
    - N: int, the number of time steps
    - pu: float, the probability of an up movement
    - pd_: float, the probability of a down movement
    - div: float, the dividend yield
    - sigma: float, the volatility of the underlying asset
    - num_of_paths: int, the number of paths to simulate
    - call: bool, whether the option is a call or a put
    - american: bool, whether the option is American or European
    - verbose: bool, the verbosity of the output
    - display: bool, whether to display the asset price paths
    """
    super().__init__(S, K, r, T, N, pu, pd_, div, sigma, call, american, verbose)
    self.num_of_paths = num_of_paths
    self.verbose = verbose
    self.show = display
  
  def simulate_price(self):
    """ Function to simulate the price of the underlying asset"""
    if self.verbose: print("Simulating the price of the underlying asset")
    S = np.zeros((self.N + 1, self.num_of_paths))
    S[0] = self.S
    for t in range(1, self.N+1):
      drift = (self.r - self.div - 0.5 * self.sigma ** 2) * self.dt
      Z = np.random.randn(self.num_of_paths)
      diffusion = self.sigma * np.sqrt(self.dt) * Z
      S[t] = S[t-1] * np.exp(drift + diffusion)
    self.STs = S
    if self.verbose: print("Price simulation complete")
  
  def display(self):
    plt.figure(figsize=(10,10))
    sns.lineplot(data = self.STs)
    plt.title("Simulated Asset Price Paths")
    plt.xlabel("Time Steps")
    plt.ylabel("Asset Price")
    plt.gca().get_legend().remove()
    plt.show()

  def price(self):
    if self.verbose: print("Calculating the option prices using the Monte Carlo method")
    self.simulate_price()
    if self.is_european:
      if self.is_call:
        payoffs = np.maximum(self.STs[-1] - self.K, 0)
      else:
        payoffs = np.maximum(self.K - self.STs[-1], 0)
      option_prices = np.exp(-(self.r - self.div) * self.T) * payoffs
      print("Averaging the option prices")
      if self.verbose: print("Calculating the average option price")
      if self.show:
        self.display()
      return np.mean(option_prices)
    else:
      if self.is_call:
        payoffs = np.maximum(self.STs - self.K, 0)
      else:
        payoffs = np.maximum(self.K - self.STs, 0)
      for t in reversed(range(self.N)):
        in_the_money = np.where(payoffs[t] > 0)[0]
        if len(in_the_money) > 0:
          regression = np.polyfit(self.STs[t, in_the_money], payoffs[t, in_the_money] * self.df, 2)
          continuation_value = np.polyval(regression, self.STs[t, in_the_money])
          if self.is_call:
            exercise_values = np.maximum(self.STs[t, in_the_money] - self.K, 0)
          else:
            exercise_values = np.maximum(self.K - self.STs[t, in_the_money], 0)
          exercise = exercise_values > continuation_value
          payoffs[t, in_the_money[exercise]] = exercise_values[exercise]
          payoffs *= self.df
      if self.verbose: print("Calculating the average option price")
      if self.show:
        self.display()
      return np.mean(payoffs[0])   

class _FDUS(_FDEU):
  """Internal class to implement the Explicit Finite Difference Method for American options"""

  def traverse_grid(self):
    if self.verbose: print("Traversing the grid for U.S. option")
    for j in reversed(range(self.N)): 
      for i in range(1, self.M): 
          calculated_value = self.a[i] * self.grid[i-1, j+1] + self.b[i] * self.grid[i, j+1] + self.c[i] * self.grid[i+1, j+1]
          intrinsic_value = max(self.boundary_conds[i] - self.K, 0) if self.is_call else max(self.K - self.boundary_conds[i], 0)
          self.grid[i, j] = max(calculated_value, intrinsic_value)
    if self.verbose: print("Grid traversed with early exercise considered")

class _FDImpUS(_FDImpEU):
  def traverse_grid(self):
    """ Solve using linear systems of equations """
    if self.verbose: print("Traversing the grid for U.S. option")
    P, L, U = linalg.lu(self.coeffs)
    aux = np.zeros(self.M-1)
    for i in reversed(range(self.N)):
      aux[0] = np.dot(-self.a[1], self.grid[0, i])
      x1 = linalg.solve(L, self.grid[1:self.M, i+1]+aux)
      x2 = linalg.solve(U, x1)
      self.grid[1:self.M, i] = x2
      intrinsic_values = np.maximum(self.boundary_conds[1:self.M] - self.K, 0) if self.is_call else np.maximum(self.K - self.boundary_conds[1:self.M], 0)
      self.grid[1:self.M, i] = np.maximum(self.grid[1:self.M, i], intrinsic_values)
    if self.verbose: print("Grid traversed with early exercise considered")

class OptionPricing(object):
  def __init__(self, 
               ticker: str,
               call: bool,
               american: bool,
               risk_free_rate: Union[str, float],
               identification: Union[str, int],
               risk_free_rate_approximation: str = 'last_observed',
               volatility: float = None,
               use_yfinance_volatility: bool = True,
               optimize_garch: bool = False,
               optimization_rounds: int = 100,
               verbose: bool = False,
               risk_free_rate_approximation_parameters: dict = {}
               ):
    """
    Class to price options using different methods.
    Inputs:
    - ticker: str, can be either the name of the company or the ticker
    - call: bool, whether the option is a call or a put
    - american: bool, whether the option is American or European
    - risk_free_rate: either a string representing the horizon or a value
    - identification: either a string representing the contract symbol or an integer representing the index of the option within the chain
    - risk_free_rate_approximation: str, the method to approximate the risk-free rate, can be 'last_observed', 'SMA', 'EMA', 'SARIMAX'
    - volatility: float, the volatility of the underlying asset
    - use_yfinance_volatility: bool, whether to use the volatility from Yahoo Finance if no volatility is provided. If False, a GARCH model will be used
    - optimize_garch: bool, whether to optimize the GARCH model for the volatility
    - optimization_rounds: int, the number of rounds to optimize the GARCH model
    - verbose: bool, whether to print the steps of the process
    - risk_free_rate_approximation_parameters: dict, the parameters to be passed to the risk-free rate approximation method
    """
    if type(ticker) != str:
      raise ValueError('ticker must be a string')
    if type(call) != bool:
      raise ValueError('call must be a boolean')
    if type(american) != bool:
      raise ValueError('american must be a boolean')
    if (type(volatility) != float) and (volatility is not None) and (type(volatility) != int):
      raise ValueError('volatility must be a number')
    if type(use_yfinance_volatility) != bool:
      raise ValueError('use_yfinance_volatility must be a boolean')
    if type(optimize_garch) != bool:
      raise ValueError('optimize_garch must be a boolean')
    if type(optimization_rounds) != int:
      raise ValueError('optimization_rounds must be an integer')
    if type(verbose) != bool:
      raise ValueError('verbose must be a boolean')
    self.fetcher = Fetcher(ticker)
    self.dataframe, self.yf_ticker = self.fetcher.df, self.fetcher.yf_stock
    if type(identification) == str:
      self.option_chain = self.yf_ticker.option_chain().calls if call else self.yf_ticker.option_chain().puts
      self.option = self.option_chain[self.option_chain['contractSymbol'] == identification]
    elif type(identification) == int:
      self.option = self.yf_ticker.option_chain().calls.iloc[identification] if call else self.yf_ticker.option_chain().puts.iloc[identification]
    else:
      raise ValueError('identification must be either a string or an integer')
    self.verbose = verbose
    self.date = self._calculate_expiration()
    self.is_call = call
    self.dividend_yield = self.fetcher.get_dividend_yield()
    if type(risk_free_rate) == str:
      self.historical = self.fetcher.get_risk_free_rate(risk_free_rate)['Close']
      if risk_free_rate_approximation == 'last_observed':
        self.risk_free_rate = self.historical.iloc[-1]
      elif risk_free_rate_approximation == 'SMA':
        if 'window' not in risk_free_rate_approximation_parameters.keys():
          print("Window not passed, using default value of 20")
        if 'num_of_steps' not in risk_free_rate_approximation_parameters.keys():
          print("Number of steps not passed, using default value of 1")
        window = risk_free_rate_approximation_parameters.get('window', 20)
        num_of_steps = risk_free_rate_approximation_parameters.get('num_of_steps', 1)
        self.risk_free_rate = _SMA(self.historical, window).predict_val(num_of_steps)
      elif risk_free_rate_approximation == 'EMA':
        if 'window' not in risk_free_rate_approximation_parameters.keys():
          print("Window not passed, using default value of 20")
        if 'num_of_steps' not in risk_free_rate_approximation_parameters.keys():
          print("Number of steps not passed, using default value of 1")
        window = risk_free_rate_approximation_parameters.get('window', 20)
        num_of_steps = risk_free_rate_approximation_parameters.get('num_of_steps', 1)
        self.risk_free_rate = _EMA(self.historical, window).predict_val(num_of_steps)
      elif risk_free_rate_approximation == 'SARIMAX':
        if verbose:
          print("The SARIMAX method accepts the following optional parameters:\n - exog: pd.Series, exogenous variable to use in the model")
          print("- forecast_steps: int, number of forecast steps")
          print("- optimize: bool, whether to optimize the hyperparameters")
          print("- optimization_rounds: int, number of rounds to optimize the hyperparameters")
          print("- test_size: float, size of the test set")
          print("- random_state: int, random state for reproducibility")
          print("- p: int, AR order")
          print("- d: int, differencing order")
          print("- q: int, MA order")
          print("- P: int, seasonal AR order")
          print("- D: int, seasonal differencing order")
          print("- Q: int, seasonal MA order")
          print("- s: int, seasonal period")
          print("- trend: str, trend parameter")
          print("- measure_error: bool, whether to measure the error")
          print("- time_varying_regression: bool, whether to use time varying regression")
          print("- mle_regression: bool, whether to use MLE regression")
          print("- simple_differencing: bool, whether to use simple differencing")
          print("- enforce_stationarity: bool, whether to enforce stationarity")
          print("- enforce_invertibility: bool, whether to enforce invertibility")
          print("- hamilton_representation: bool, whether to use Hamilton representation")
          print("- concentrate_scale: bool, whether to concentrate scale")
          print("- trend_offset: int, trend offset")
          print("- use_exact_diffuse: bool, whether to use exact diffuse")
        if 'forecast_steps' in risk_free_rate_approximation_parameters.keys():
          steps = risk_free_rate_approximation_parameters['forecast_steps']
          risk_free_rate_approximation_parameters.pop('forecast_steps')
        if steps is None or not isinstance(steps, int):
          steps = 1
        model = _SARIMAX_model(self.historical, **risk_free_rate_approximation_parameters)
        self.risk_free_rate = np.mean(model.forecast(steps))

    elif type(risk_free_rate) == float or type(risk_free_rate) == int:
      self.risk_free_rate = risk_free_rate
    else:
      raise ValueError('risk_free_rate must be either a string representing the horizon or a value')
    self.r = self.risk_free_rate
    self.S = self.dataframe['Close'].iloc[-1]
    self.T = (self.date - pd.to_datetime('today')).days / 365
    self.K = self.option['strike']
    self.american = american
    if use_yfinance_volatility:
      self.sigma = self.option['impliedVolatility']
    else:
      if optimize_garch:
        garch_params = self.optimize_garch_model(optimization_rounds)
        self.build_garch_model(**garch_params)
      else:
        self.build_garch_model('GARCH', 'Constant', 1, 1, 'normal', 0, 2, None, None) 
      res = self.garch_model.fit(disp='off')
      sigma = res.conditional_volatility
      self.sigma = sigma[-1]
    if volatility is not None:
      self.sigma = volatility

  def _from_name_to_datestr(self, s: str) -> str:
    """Helper function to convert the contract symbol to a date string"""
    if self.verbose:
      print(f"Processing {s}")
    match = re.search(r'[A-Za-z](\d{2})(\d{2})(\d{2})[CP]', s)
    if match:
        year, month, day = match.groups()
        return f"20{year}-{month}-{day}"
    else:
        return "No date found"
  
  def _calculate_expiration(self) -> pd.Timestamp:
    """Function to calculate the expiration date of the option"""
    if self.verbose:
      print("Calculating expiration date")
    to_process = self.option['contractSymbol']
    date_str = self._from_name_to_datestr(to_process)
    date = pd.to_datetime(date_str)
    return date 
  
  def calculate_option_price(self, method: str, describe: bool = False, **kwargs) -> float:
    """
    Main wrapper function to calculate the price of an option using the specified method.
    Inputs:
      - method: str, the method to use for pricing
      - describe: bool, whether to describe the method
      - kwargs: additional arguments to be passed to the method
    
    Accepted methods:
    - Binomial Tree Methods: Binomial Tree, Binomial Lattice Tree, Cox-Ross-Rubinstein, Leisen-Reimer Tree
    - Trinomial Tree Methods: Trinomial Tree, Trinomial Lattice Tree
    - Finite Difference Methods: Explicit Finite Difference, Implicit Finite Difference, Crank-Nicolson Finite Difference
    """
    method = method.replace(' ', '').lower()
    if method in ('blackscholes', 'black-scholes', 'bs', 'black_scholes'):
      if self.american:
        raise MethodError("The Black-Scholes method is only available for European options")
      if describe:
        print("The Black-Scholes Method requires no additional values")
      price = self.black_scholes()
      return price
    
    if method in ('montecarlo', 'monte-carlo', 'mc', 'monte_carlo'):
      if 'num_of_paths' not in kwargs:
        print("num_of_paths not passed, using default value of 1000")
      if 'N' not in kwargs:
        print("N not passed, using default value of 1000")
      if 'display' not in kwargs:
        print("display not passed, using default value of False")
      if describe:
        print("The Monte Carlo Method requires three optional value,")
        print("num_of_paths, which is the number of paths to simulate")
        print("N, which is the number of time steps")
        print("display, which is whether to display the asset price paths")
      num_of_paths = kwargs.get('num_of_paths', 1000)
      N = kwargs.get('N', 1000)
      display = kwargs.get('display', False)
      if type(N) != int:
        raise ValueError("N must be an integer")
      if type(num_of_paths) != int:
        raise ValueError("num_of_paths must be an integer")
      if type(display) != bool:
        raise ValueError("display must be a boolean")
      price = self.montecarlo(num_of_paths, N, display)
      return price

    if method in ('binomialtree', 'binomial', 'binomtree', 'binom'): #Check
      if 'N' not in kwargs:
        print("N not passed, using default value of 1000")
      if 'pd_' not in kwargs:
        print("pd_ not passed, using default value of 0.01")
      if 'pu' not in kwargs:
        print("pu not passed, using default value of 0.01")
      if describe:
        print("The Binomial Tree Method requires three optional values,")
        print("N, which is the number of time steps")
        print("pd_, which is the probability at the down state")
        print("pu, which is the probability at the up state")
      N = kwargs.get('N', 1000)
      pd_ = kwargs.get('pd_', 1e-2)
      pu = kwargs.get('pu', 1e-2)
      if self.american:
        price = self.binom_american(N, pd_, pu)
      else:
        price = self.binom_european(N, pd_, pu)
      return price
    
    if method in ('binomiallattice', 'binomial_lattice', 'binomlattice', 'binom_lattice'): #Check
      if 'N' not in kwargs:
        print("N not passed, using default value of 2")
      if describe:
        print("The Binomial Lattice Method requires one optional values,")
        print("N, which is the number of time steps")
      N = kwargs.get('N', 2)
      price = self.binom_lattice(N)
      return price
    
    if method in ('cox-ross-rubinstein', 'cox', 'coxrossrubinstein', 'crr'): #Check
      if 'N' not in kwargs:
        print("N not passed, using default value of 10")
      if describe:
        print("The Cox-Ross-Rubinstein Method requires one optional values,")
        print("N, which is the number of time steps")
      N = kwargs.get('N', 10)
      price = self.cox_ross_rubinstein(N)
      return price
    
    if method in ('leisen-reimertree', 'lrtree', 'lr'): #Check
      if 'N' not in kwargs:
        print("N not passed, using default value of 5")
      if describe:
        print("The Leisen-Reimer Tree Method requires one optional value,")
        print("N, which is the number of time steps")
      N = kwargs.get('N', 5)
      price = self.lr_tree(N)
      return price
    
    if method in ('trinomialtree', 'trinomial', 'trinom'): #Check
      if 'N' not in kwargs:
        print("N not passed, using default value of 1000")
      N = kwargs.get('N', 1000)
      if describe:
        print("The Trinomial Tree Method requires one optional value,")
        print("N, which is the number of time steps")
      price = self.trinom(N)
      return price
    
    if method in ('trinomiallattice', 'trinomial_lattice', 'trinomlattice', 'trinom_lattice'): #Check
      if 'N' not in kwargs:
        print("N not passed, using default value of 2")
      N = kwargs.get('N', 2)
      if describe:
        print("The Trinomial Lattice Method requires three optional values,")
        print("N, which is the number of time steps")
      price = self.trinom_lattice(N)
      return price
    
    if method in ('finitedifference', 'finite_difference'):
      raise MethodError("Please specify between Explicit Finite Difference, Implicit Finite Difference and Crank-Nicolson Finite Difference")
    
    if method in ('explicitfinitedifference', 'explicit', 'explicitdifference', 'explicitfinite', 'explicit_difference'):
      if 'Smax' not in kwargs:
        print("Smax not passed, using default value of 1000")
      if 'M' not in kwargs:
        print("M not passed, using default value of 50")
      if 'N' not in kwargs:
        print("N not passed, using default value of 50")
      Smax = kwargs.get('Smax', 1000)
      M = kwargs.get('M', 50)
      N = kwargs.get('N', 50)
      if describe:
        print("The Explicit Finite Difference Method requires three optional values,")
        print("Smax, which is the maximum value of the underlying asset")
        print("M, which is the number of price steps")
        print("N, which is the number of time steps")
      if self.american:
        price = self.explicit_american(Smax, M, N)
      else:
        price = self.explicit_european(Smax, M, N)
      return price
    
    if method in ('implicitfinitedifference', 'implicit', 'implicitdifference', 'implicitfinite'):
      if 'Smax' not in kwargs:
        print("Smax not passed, using default value of 1000")
      if 'M' not in kwargs:
        print("M not passed, using default value of 1000")
      if 'N' not in kwargs:
        print("N not passed, using default value of 1000")
      Smax = kwargs.get('Smax', 1000)
      M = kwargs.get('M', 1000)
      N = kwargs.get('N', 1000)
      if describe:
        print("The Implicit Finite Difference Method requires three optional values,")
        print("Smax, which is the maximum value of the underlying asset")
        print("M, which is the number of price steps")
        print("N, which is the number of time steps")
      if self.american:
        price = self.implicit_american(Smax, M, N)
      else:
        price = self.implicit_european(Smax, M, N)
      return price
    
    if method in ('crank-nicolsonfinitedifference', 'crank-nicolson', 'cranknicolson', 'crank-nicolsonfinite', 'crank-nicolsondifference', 'cranknicolsonfinite', 'cranknicolsondifference', 'crank_nicolson', 'crank_nicolson_finite', 'crank_nicolson_difference','crank_nicolson_finite_difference', 'cnfd', 'cn'):
      if self.american:
        if 'Smax' not in kwargs:
          print("Smax not passed, using default value of 1000")
        if 'M' not in kwargs:
          print("M not passed, using default value of 100")
        if 'N' not in kwargs:
          print("N not passed, using default value of 100")
        if 'omega' not in kwargs:
          print("omega not passed, using default value of 1.2")
        if 'tol' not in kwargs:
          print("tol not passed, using default value of 1e-2")
        Smax = kwargs.get('Smax', 1000)
        M = kwargs.get('M', 100)
        N = kwargs.get('N', 100)
        omega = kwargs.get('omega', 1.2)
        tol = kwargs.get('tol', 1e-2)
        if describe:
          print("The Crank-Nicolson Finite Difference Method for american options requires five optional values,")
          print("Smax, which is the maximum value of the underlying asset")
          print("M, which is the number of price steps")
          print("N, which is the number of time steps")
          print("omega, which is the relaxation factor")
          print("tol, which is the tolerance")
        price = self.crank_nicolson_american(Smax, M, N, omega, tol)
      else:
        if 'Smax' not in kwargs:
          print("Smax not passed, using default value of 1000")
        if 'M' not in kwargs:
          print("M not passed, using default value of 100")
        if 'N' not in kwargs:
          print("N not passed, using default value of 100")
        Smax = kwargs.get('Smax', 1000)
        M = kwargs.get('M', 100)
        N = kwargs.get('N', 100)
        if describe:
          print("The Crank-Nicolson Finite Difference Method for european options requires three optional values,")
          print("Smax, which is the maximum value of the underlying asset")
          print("M, which is the number of price steps")
          print("N, which is the number of time steps")
        price = self.crank_nicolson_european(Smax, M, N)
      return price
    raise MethodError(f"Method {method} not found")
  
  def montecarlo(self, num_of_paths: int = 1000, N: int = 1000, display = False, pu = 0, pd_ = 0) -> float:
    mc = _MonteCarlo(self.S, self.K, self.risk_free_rate, self.T, N, pu, pd_, self.dividend_yield, self.sigma, num_of_paths, self.is_call, self.american, self.verbose, display)
    return mc.price()

  def black_scholes(self, N = 100, pu = 0, pd_ = 0) -> float:
    bs = _BlackScholes(self.S, self.K, self.risk_free_rate, self.T, N, pu, pd_, self.dividend_yield, self.sigma, self.is_call, self.american, self.verbose)
    return bs.price()

  def crank_nicolson_american(self, Smax: int = 1000, M: int = 100, N: int = 100, omega: float = 1.2, tol: float = 1e-3) -> float:
    us = _FDCNUS(self.S, self.K, self.r, self.T, self.sigma, Smax, M, N, omega, tol, self.is_call, self.verbose)
    return us.price()

  def crank_nicolson_european(self, Smax: int = 1000, M: int = 100, N: int = 100) -> float:
    eu = _FDCN(self.S, self.K, self.r, self.T, self.sigma, Smax, M, N, self.is_call, self.verbose)
    return eu.price()

  def implicit_american(self, Smax: int = 1000, M: int = 1000, N: int = 1000) -> float:
    us = _FDImpUS(self.S, self.K, self.r, self.T, self.sigma, Smax, M, N, self.is_call, self.verbose)
    return us.price()

  def implicit_european(self, Smax: int = 1000, M: int = 1000, N: int = 1000) -> float: 
    eu = _FDImpEU(self.S, self.K, self.r, self.T, self.sigma, Smax, M, N, self.is_call, self.verbose)
    return eu.price() 

  def explicit_american(self, Smax: int = 1000, M: int = 1000, N: int = 1000) -> float:
    us = _FDUS(self.S, self.K, self.r, self.T, self.sigma, Smax, M, N, self.is_call, self.verbose)
    return us.price()

  def explicit_european(self, Smax: int = 1000, M: int = 1000, N: int = 1000) -> float:
    eu = _FDEU(self.S, self.K, self.r, self.T, self.sigma, Smax, M, N, self.is_call, self.verbose)
    return eu.price() 

  def trinom_lattice(self, N: int = 1000, pd_: float = 0, pu: float = 0) -> float:
    option = _TrinomialLattice(self.S, self.K, self.risk_free_rate, self.T, N, pu, pd_, self.dividend_yield, self.sigma, self.is_call, self.american, self.verbose)
    return option.price()

  def binom_lattice(self, N: int = 1000, pd_: float = 0, pu: float = 0) -> float:
    option = _BinomialCRRLattice(self.S, self.K, self.risk_free_rate, self.T, N, pu, pd_, self.dividend_yield, self.sigma, self.is_call, self.american, self.verbose)
    return option.price()

  def trinom(self, N: int = 1000, pd_: float = 0, pu: float = 0) -> float:
    option = _TrinomialTree(self.S, self.K, self.risk_free_rate, self.T, N, pu, pd_, self.dividend_yield, self.sigma, self.is_call, self.american, self.verbose)
    return option.price()
  
  def lr_tree(self, N: int = 1000, pd_: float = 0, pu: float = 0) -> float:
    opt = _BinomialLR(self.S, self.K, self.risk_free_rate, self.T, N, pu, pd_, self.dividend_yield, self.sigma, self.is_call, self.american, self.verbose)
    return opt.price()

  def cox_ross_rubinstein(self, N: int = 1000, pd_: float = 0, pu: float = 0) -> float:
    opt = _BinomialCRR(self.S, self.K, self.risk_free_rate, self.T, N, pu, pd_, self.dividend_yield, self.sigma, self.is_call, self.american, self.verbose)
    return opt.price()

  def binom_american(self, N: int = 1000, pd_: float = 0, pu: float = 0) -> float:
    us = _USBinomialTree(self.S, self.K, self.risk_free_rate, self.T, N, pu, pd_, self.dividend_yield, self.sigma, self.is_call, self.american, self.verbose)
    return us.price()
  
  def binom_european(self, N: int, pd_: float, pu: float) -> float:
    eu = _BinomialTree(self.S, self.K, self.risk_free_rate, self.T, N, pu, pd_, self.dividend_yield, self.sigma, self.is_call, self.american, self.verbose)
    return eu.price()
  
  def build_garch_model(self, vol, mean, p, q, dist, o, power, hold_back, rescale):
    if self.verbose: print("Building GARCH model")
    target_val = self.dataframe['Close'].pct_change().dropna() * 100
    self.garch_model = arch.arch_model(target_val, vol = vol, mean = mean, p=p, q=q, dist=dist, o=o, power=power, hold_back=hold_back, rescale=rescale)
    if self.verbose: print("GARCH model built")

  def optimize_garch_model(self, optimization_trials) -> dict:
    """
    This function is designed to optimize the GARCH model on the returns of the stock.
    Optuna is used to optimize the parameters of the GARCH model.
    Inputs:
      - optimization_trials: int, the number of trials to run
    Warning: This function is experimental
    """
    if self.verbose: print("Optimizing GARCH model")
    def objective(trial):
        vol =  trial.suggest_categorical('vol', ['GARCH', 'ARCH', 'EGARCH', 'FIGARCH', 'APARCH', 'HARCH'])
        p = trial.suggest_int('p', 0, 10)
        q = trial.suggest_int('q', 0, 10)
        mean = trial.suggest_categorical('mean', ['Constant', 'Zero', 'LS', 'AR', 'ARX', 'HAR', 'HARX', 'constant', 'zero'])
        dist = trial.suggest_categorical('dist', ['normal', 'gaussian', 't', 'studentst', 'skewstudent', 'skewt', 'ged', 'generalized error'])
        o = trial.suggest_int('o', 0, 10)
        power = trial.suggest_float('power', 0, 10)
        hold_back = trial.suggest_int('hold_back', 0, 10)
        if hold_back == 0: hold_back = None
        rescale = trial.suggest_categorical('rescale', [True, False, None])
        if self.verbose: print(f"Optimizing with parameters {vol}, {mean}, {p}, {q}, {dist}, {o}, {power}, {hold_back}, {rescale}")
        try:
            self.build_garch_model(vol, mean, p, q, dist, o, power, hold_back, rescale)
            res = self.garch_model.fit(disp='off')
            return res.aic
        except Exception as e:
            print(f"Error with parameters {p}, {q}, {vol}, {mean}, {dist}, {o}, {power}, {hold_back}, {rescale}: {e}")
            return float('inf')
    print("Warning: This function is experimental, and may throw errors in the optimization process. Use at your own risk.")
    study = optuna.create_study(direction='minimize')
    study.optimize(objective, n_trials=optimization_trials)
    if self.verbose: print("Optimization complete")
    return study.best_params
  
  def __repr__(self) -> str:
    return f"Option of {self.yf_ticker.ticker}\nStrike Price: {self.K}\nRisk-Free Rate: {self.risk_free_rate}\nExpiration Date: {self.date}\nDividend Yield: {self.dividend_yield}\nPrice of the stock: {self.S}\nVolatility: {self.sigma}\nIs Call: {self.is_call}\nIs an american option: {self.american}\n"



"""
This module contains the IndicatorCalculator class which is designed to do multiple manipulation with financial data
"""
import pandas as pd
import numpy as np
import pandas_datareader.data as web
from .datafetcher import Fetcher

class IndicatorCalculator:
    """
    A utility class for performing various technical analysis calculations on stock market data.
    
    This class supports operations such as calculating moving averages, exponential moving averages,
    Bollinger Bands, Relative Strength Index (RSI), Moving Average Convergence Divergence (MACD),
    and many other technical indicators. It can work with data either fetched directly using a ticker symbol
    or with an externally supplied pandas DataFrame.

    Attributes:
        - ticker (str, optional): The ticker symbol of the stock to fetch data for. If specified, data
                                  will be fetched using the Fetcher class.
        - dataframe (pd.DataFrame, optional): An externally supplied pandas DataFrame containing stock
                                              market data. Expected to have columns like 'Close', 'High',
                                              'Low', and 'Volume'.
        - verbose (int, optional): The verbosity level of the class, with 0 meaning no output and higher
                                   values meaning higher verbosity. Maximum verbosity is 2.
        - df (pd.DataFrame): The primary DataFrame used for calculations, derived either from the fetched
                             data using the ticker symbol or directly supplied as an argument.
        - yf_stock (yf.Ticker, optional): The yfinance Ticker object for the specified ticker symbol,
                                          available if a ticker symbol was provided.

    Methods:
        This class provides a variety of methods for technical analysis calculations, including:
        - exponential_weights: Applies exponential weights to the 'Close' column.
        - adjust_inflation: Adjusts the 'Close' column for inflation using CPI data.
        - calculate_returns: Calculates the simple returns of the 'Close' column.
        - calculate_RSI: Calculates the Relative Strength Index.
        - calculate_MACD: Calculates the Moving Average Convergence Divergence.
        - calculate_bollinger_bands: Calculates Bollinger Bands.
        - calculate_EMA: Calculates the Exponential Moving Average.
        - calculate_SMA: Calculates the Simple Moving Average.
        - calculate_KAMA: Calculates Kaufman's Adaptive Moving Average.
        - calculate_on_balance_volume: Calculates On-Balance Volume.
        - calculate_ATR: Calculates the Average True Range.
        - calculate_ADI: Calculates the Accumulation/Distribution Index.
        - (and many more methods for various technical indicators...)

    Usage:
        To use this class, either provide a ticker symbol to fetch data or supply a pandas DataFrame directly.
        Once instantiated, call the desired method(s) to perform technical analysis calculations.

    Example:
        indicators = IndicatorCalculator(ticker='AAPL')
        df_with_rsi = indicators.calculate_RSI()
        df_with_macd = indicators.calculate_MACD()

        # Or with an external DataFrame
        external_df = pd.DataFrame(...)  # Some DataFrame with market data, must have the columns each method expects
        indicators = IndicatorCalculator(dataframe=external_df)
        df_with_bollinger = indicators.calculate_bollinger_bands()
    """

    def __init__(self, 
                 ticker:str = None, 
                 dataframe:pd.DataFrame = None,
                 verbose: int = 0):
        if ticker is not None:
            self.fetch = Fetcher(ticker)
            self.df, self.yf_stock = self.fetch.df, self.fetch.yf_stock
        elif dataframe is not None:
            self.df = dataframe
        else:
            print("Warning: No data provided, dataframe will need to be supplied for each method call")
        if verbose < 0 or verbose > 2 or type(verbose) != int:
            raise ValueError("Invalid verbosity level. Must be an integer between 0 and 2.")
        self.verb = verbose
    
    def exponential_weights(self, 
                            span:int, 
                            df:pd.DataFrame = None) -> pd.DataFrame:
        """Modifies the 'Close' column of the dataframe by applying exponential weights to it
        Inputs:
        - span: int: The span of the exponential weights
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close' column - if None, the class dataframe is used
        """
        if self.verb == 2:
            print(f"Applying exponential weights with span {span}")
        if df is None:
            df = self.df
        weights = np.exp(-np.log(2)*np.arange(len(df))/span)[::-1]
        weighted_close = df['Close'] * weights / weights.max()
        df['Close'] = weighted_close
        if self.verb > 0:
            print("Exponential weights applied")
        return df
    
    def adjust_inflation(self, 
                         df:pd.DataFrame = None) -> pd.DataFrame:
        """Adjusts the 'Close' column of the dataframe for inflation
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close' column - if None, the class dataframe is used
        """
        if self.verb == 2:
            print("Adjusting the 'Close' column for inflation")
        if df is None:
            df = self.df.copy()
        df.index = df.index.tz_localize(None)
        first_date = df.index.min()
        last_date = df.index.max()
        cpi_series_id = 'CPIAUCSL'
        cpi_data = web.DataReader(cpi_series_id, 'fred', first_date, last_date)
        cpi_data.rename(columns = {'CPIAUCSL':'CPI'}, inplace = True)
        base_cpi = cpi_data['CPI'].iloc[-1] 
        cpi_all_data = cpi_data.reindex(df.index, method = 'ffill')
        cpi_all_data = cpi_all_data.fillna(method = 'bfill')
        refactor = base_cpi / cpi_all_data['CPI']
        df['Refactored'] = df['Close'] * refactor
        if self.verb > 0:
            print("Inflation adjustment complete")
        return df

    def calculate_returns(self, 
                          df:pd.DataFrame = None) -> pd.DataFrame:
        """Calculates the returns of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close' column - if None, the class dataframe is used
        """
        if self.verb == 2:
            print("Calculating returns")
        if df is None:
            df = self.df
        df['Return'] = df['Close'].pct_change()
        if self.verb > 0:
            print("Returns calculated")
        return df
    
    def calculate_RSI(self, 
                      df:pd.DataFrame = None, 
                      period:int = 14) -> pd.DataFrame:
        """Calculates the Relative Strength Index of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close' column - if None, the class dataframe is used
        - period: int: The period to be used for the RSI calculation - default is 14"""
        if self.verb == 2:
            print(f"Calculating RSI with period: {period}")
        if df is None:
            df = self.df
        delta = df['Close'].diff()
        gain = (delta.where(delta > 0, 0)).rolling(period).mean()
        loss = (-delta.where(delta < 0, 0)).rolling(period).mean()
        RS = gain / loss
        RSI = 100 - (100 / (1 + RS))
        df['RSI'] = RSI
        if self.verb > 0:
            print("RSI calculated")
        return df
    
    def calculate_MACD(self, 
                       df:pd.DataFrame = None, 
                       short_window:int = 12, 
                       long_window:int = 26) -> pd.DataFrame:
        """Calculates the Moving Average Convergence Divergence of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close' column - if None, the class dataframe is used
        - short_window: int: The period to be used for the short EMA - default is 12
        - long_window: int: The period to be used for the long EMA - default is 26
        """
        if self.verb == 2:
            print(f"Calculating MACD with short window: {short_window} and long window: {long_window}")
        if df is None:
            df = self.df
        short_ema = df['Close'].ewm(span = short_window, adjust = False).mean()
        long_ema = df['Close'].ewm(span = long_window, adjust = False).mean()
        df['MACD'] = short_ema - long_ema
        df['Signal'] = df['MACD'].ewm(span = 9, adjust = False).mean()
        if self.verb > 0:
            print("MACD calculated")
        return df
    
    def calculate_bollinger_bands(self, 
                                  df:pd.DataFrame = None, 
                                  window:int = 20) -> pd.DataFrame:
        """Calculates the Bollinger Bands of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close' column - if None, the class dataframe is used
        - window: int: The period to be used for the Bollinger Bands calculation - default is 20"""
        if self.verb == 2:
            print(f"Calculating Bollinger Bands with window: {window}")
        if df is None:
            df = self.df
        rolling_mean = df['Close'].rolling(window).mean()
        rolling_std = df['Close'].rolling(window).std()
        df['Upper'] = rolling_mean + (rolling_std * 2)
        df['Lower'] = rolling_mean - (rolling_std * 2)
        if self.verb > 0:
            print("Bollinger Bands calculated")
        return df

    def calculate_EMA(self, 
                      df:pd.DataFrame = None, 
                      window:int = 20) -> pd.DataFrame:
        """Calculates the Exponential Moving Average of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close' column - if None, the class dataframe is used
        - window: int: The period to be used for the EMA calculation - default is 20"""
        if self.verb == 2:
            print(f"Calculating EMA with window: {window}")
        if df is None:
            df = self.df
        df['EMA'] = df['Close'].ewm(span = window, adjust = False).mean()
        if self.verb > 0:
            print("EMA calculated")
        return df
    
    def calculate_SMA(self, 
                      df:pd.DataFrame = None, 
                      window:int = 20) -> pd.DataFrame:
        """Calculates the Simple Moving Average of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close' column - if None, the class dataframe is used
        - window: int: The period to be used for the SMA calculation - default is 20"""
        if self.verb == 2:
            print(f"Calculating SMA with window: {window}")
        if df is None:
            df = self.df
        df['SMA'] = df['Close'].rolling(window).mean()
        if self.verb > 0:
            print("SMA calculated")
        return df
    
    def calculate_KAMA(self, 
                       df:pd.DataFrame = None, 
                       window:int = 20, 
                       fast:int=2, 
                       slow: int =30) -> pd.DataFrame:
        """Calculates the Kaufman's Adaptive Moving Average of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close' column - if None, the class dataframe is used
        - window: int: The period to be used for the KAMA calculation - default is 20
        - fast: int: The period to be used for the fast smoothing constant - default is 2
        - slow: int: The period to be used for the slow smoothing constant - default is 30"""
        if self.verb == 2:
            print(f"Calculating KAMA with window: {window}, fast: {fast}, slow: {slow}")
        if df is None:
            df = self.df
        change = df['Close'].diff(window).abs()
        volatility = df['Close'].diff().abs().rolling(window=window).sum()
        ER = change / volatility
        fastSC = 2 / (fast + 1)
        slowSC = 2 / (slow + 1)
        SC = (ER * (fastSC - slowSC) + slowSC) ** 2
        KAMA = pd.Series(index=df.index, data=np.NaN)
        KAMA.iloc[window] = df['Close'].iloc[window]
        for i in range(window + 1, len(df)):
            KAMA.iloc[i] = KAMA.iloc[i - 1] + SC.iloc[i] * (df['Close'].iloc[i] - KAMA.iloc[i - 1])
        df['KAMA'] = KAMA
        if self.verb > 0:
            print("KAMA calculated")
        return df
    
    def calculate_on_balance_volume(self, 
                                    df:pd.DataFrame = None) -> pd.DataFrame:
        """Calculates the On Balance Volume of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close' and 'Volume' column - if None, the class dataframe is used"""
        if self.verb == 2:
            print("Calculating On Balance Volume")
        if df is None:
            df = self.df
        df['OBV'] = np.where(df['Close'] > df['Close'].shift(1), df['Volume'], np.where(df['Close'] < df['Close'].shift(1), -df['Volume'], 0)).cumsum()
        if self.verb > 0:
            print("On Balance Volume calculated")
        return df

    def calculate_ATR(self, 
                      df:pd.DataFrame = None, 
                      window:int = 14) -> pd.DataFrame:
        """Calculates the Average True Range of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close', 'High' and 'Low' column - if None, the class dataframe is used
        - window: int: The period to be used for the ATR calculation - default is 14"""
        if self.verb == 2:
            print(f"Calculating ATR with window: {window}")
        if df is None:
            df = self.df
        df['H-L'] = abs(df['High'] - df['Low'])
        df['H-PC'] = abs(df['High'] - df['Close'].shift(1))
        df['L-PC'] = abs(df['Low'] - df['Close'].shift(1))
        df['TR'] = df[['H-L', 'H-PC', 'L-PC']].max(axis = 1, skipna = False)
        df['ATR'] = df['TR'].rolling(window).mean()
        if self.verb > 0:
            print("ATR calculated")
        return df
    
    def calculate_ADI(self, 
                      df: pd.DataFrame = None) -> pd.DataFrame:
        """Calculates the Accumulation/Distribution Index of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close', 'High', 'Low' and 'Volume' column - if None, the class dataframe is used"""
        if self.verb == 2:
            print("Calculating ADI")
        if df is None:
            df = self.df
        clv = ((df['Close'] - df['Low']) - (df['High'] - df['Close'])) 
        range_span = (df['High'] - df['Low'])
        df['MFM'] = np.where(range_span == 0, 0, clv / range_span)  
        df['MFMV'] = df['MFM'] * df['Volume']
        df['ADI'] = df['MFMV'].cumsum()
        if self.verb > 0:
            print("ADI calculated")
        return df


    def calculate_parabolic_sar(self, 
                                df: pd.DataFrame = None, 
                                acceleration: float =0.02, 
                                maximum: float = 0.2) -> pd.DataFrame:
        """Calculates the Parabolic SAR of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close', 'High' and 'Low' column - if None, the class dataframe is used
        - acceleration: float: The acceleration factor to be used for the Parabolic SAR calculation - default is 0.02
        - maximum: float: The maximum acceleration factor to be used for the Parabolic SAR calculation - default is 0.2"""
        if self.verb == 2:
            print(f"Calculating Parabolic SAR with acceleration: {acceleration} and maximum: {maximum}")
        if df is None:
            df = self.df
        df['SAR'] = 0.0
        df['EP'] = 0.0
        df['AF'] = acceleration
        rising_trend = df['Close'].iloc[1] > df['Close'].iloc[0]
        if rising_trend:
            df['SAR'].iloc[0] = df['Low'].iloc[0]
            df['EP'].iloc[0] = df['High'].iloc[0]
        else:
            df['SAR'].iloc[0] = df['High'].iloc[0]
            df['EP'].iloc[0] = df['Low'].iloc[0]

        for i in range(1, len(df)):
            if (rising_trend and df['High'].iloc[i] > df['EP'].iloc[i-1]) or (not rising_trend and df['Low'].iloc[i] < df['EP'].iloc[i-1]):
                df['AF'].iloc[i] = min(df['AF'].iloc[i-1] + acceleration, maximum)
            else:
                df['AF'].iloc[i] = df['AF'].iloc[i-1]

            df['SAR'].iloc[i] = df['SAR'].iloc[i-1] + df['AF'].iloc[i-1] * (df['EP'].iloc[i-1] - df['SAR'].iloc[i-1])

            if rising_trend:
                if df['Low'].iloc[i] < df['SAR'].iloc[i]:
                    rising_trend = False
                    df['SAR'].iloc[i] = df['EP'].iloc[i-1]
                    df['EP'].iloc[i] = df['Low'].iloc[i]
                    df['AF'].iloc[i] = acceleration
                else:
                    if df['High'].iloc[i] > df['EP'].iloc[i-1]:
                        df['EP'].iloc[i] = df['High'].iloc[i]
            else:
                if df['High'].iloc[i] > df['SAR'].iloc[i]:
                    rising_trend = True
                    df['SAR'].iloc[i] = df['EP'].iloc[i-1]
                    df['EP'].iloc[i] = df['High'].iloc[i]
                    df['AF'].iloc[i] = acceleration
                else:
                    if df['Low'].iloc[i] < df['EP'].iloc[i-1]:
                        df['EP'].iloc[i] = df['Low'].iloc[i]

            if rising_trend:
                df['SAR'].iloc[i] = min(df['SAR'].iloc[i], df['Low'].iloc[i-1], df['Low'].iloc[i-2] if i >= 2 else df['SAR'].iloc[i])
            else:
                df['SAR'].iloc[i] = max(df['SAR'].iloc[i], df['High'].iloc[i-1], df['High'].iloc[i-2] if i >= 2 else df['SAR'].iloc[i])
        if self.verb > 0:
            print("Parabolic SAR calculated")
        return df
    
    def calculate_stochastic_oscillator(self, 
                                        df:pd.DataFrame = None, 
                                        window:int = 14) -> pd.DataFrame:
        """Calculates the Stochastic Oscillator of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close', 'High' and 'Low' column - if None, the class dataframe is used
        - window: int: The period to be used for the Stochastic Oscillator calculation - default is 14"""
        if self.verb == 2:
            print(f"Calculating Stochastic Oscillator with window: {window}")
        if df is None:
            df = self.df
        low_str = 'L'+str(window)
        high_str = 'H'+str(window)
        df[low_str] = df['Low'].rolling(window).min()
        df[high_str] = df['High'].rolling(window).max()
        df['%K'] = 100 * ((df['Close'] - df[low_str]) / (df[high_str] - df[low_str]))
        df['%D'] = df['%K'].rolling(window).mean()
        if self.verb > 0:
            print("Stochastic Oscillator calculated")
        return df
    
    def calculate_momentum(self, 
                           df:pd.DataFrame = None, 
                           window:int = 10) -> pd.DataFrame:
        """Calculates the Momentum of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close' column - if None, the class dataframe is used
        - window: int: The period to be used for the Momentum calculation - default is 10
        """
        if self.verb == 2:
            print(f"Calculating Momentum with window: {window}")
        if df is None:
            df = self.df
        df['Momentum'] = df['Close'] - df['Close'].shift(window)
        if self.verb > 0:
            print("Momentum calculated")
        return df
    
    def calculate_williams_R(self, 
                             df:pd.DataFrame = None, 
                             window:int = 14) -> pd.DataFrame:
        """Calculates the Williams %R of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close', 'High' and 'Low' column - if None, the class dataframe is used
        - window: int: The period to be used for the Williams %R calculation - default is 14"""
        if self.verb == 2:
            print(f"Calculating Williams %R with window: {window}")
        if df is None:
            df = self.df
        df['H'+ str(window)] = df['High'].rolling(window).max()
        df['L' + str(window)] = df['Low'].rolling(window).min()
        df['Williams %R'] = -100 * ((df['H'+ str(window)] - df['Close']) / (df['H'+ str(window)] - df['L' + str(window)]))
        if self.verb > 0:
            print("Williams %R calculated")
        return df
    
    def calculate_fibonacci_retracement(self,
                                        df:pd.DataFrame = None, 
                                        start_and_end_idxs:tuple=None, 
                                        fib_1: float = 0.236, 
                                        fib_2: float = 0.382, 
                                        fib_3: float = 0.618, 
                                        fib_4: float = 0.786) -> pd.DataFrame:
        """Calculates the Fibonacci Retracement of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close', 'High' and 'Low' column - if None, the class dataframe is used
        - start_and_end_idxs: tuple: The start and end indexes to be used for the calculation - if None, the whole dataframe is used
        - fib_1: float: The first fibonacci level - default is 0.236
        - fib_2: float: The second fibonacci level - default is 0.382
        - fib_3: float: The third fibonacci level - default is 0.618
        - fib_4: float: The fourth fibonacci level - default is 0.786"""
        if self.verb == 2:
            print(f"Calculating Fibonacci Retracement with levels: {fib_1}, {fib_2}, {fib_3}, {fib_4}")
        if df is None:
            df = self.df
        if start_and_end_idxs is None:
            period_high = df['High'].max()
            period_low = df['Low'].min()
        else:
            start_index, end_index = start_and_end_idxs
            period_high = df['High'][start_index:end_index].max()
            period_low = df['Low'][start_index:end_index].min()
        diff = period_high - period_low

        df['Fib_' + str(fib_1 * 100) + '%'] = period_high - diff * fib_1
        df['Fib_' + str(fib_2 * 100) + '%'] = period_high - diff * fib_2
        df['Fib_' + str(fib_3 * 100) + '%'] = period_high - diff * fib_3
        df['Fib_' + str(fib_4 * 100) + '%'] = period_high - diff * fib_4
        if self.verb > 0:
            print("Fibonacci Retracement calculated")
        return df
    
    def calculate_aroon_oscillator(self, 
                                   df:pd.DataFrame = None, 
                                   window:int = 25) -> pd.DataFrame:
        """Calculates the Aroon Oscillator of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close', 'High' and 'Low' column - if None, the class dataframe is used
        - window: int: The period to be used for the Aroon Oscillator calculation - default is 25"""
        if self.verb == 2:
            print(f"Calculating Aroon Oscillator with window: {window}")
        if df is None:
            df = self.df
        df['Up'] = df['High'].rolling(window).apply(lambda x: x.argmax(), raw = True) / window * 100
        df['Down'] = df['Low'].rolling(window).apply(lambda x: x.argmin(), raw = True) / window * 100
        df['Aroon Oscillator'] = df['Up'] - df['Down']
        if self.verb > 0:
            print("Aroon Oscillator calculated")
        return df
    
    def calculate_ADX(self, 
                      df:pd.DataFrame = None, 
                      window:int = 14) -> pd.DataFrame:
        """Calculates the Average Directional Index of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close', 'High' and 'Low' column - if None, the class dataframe is used
        - window: int: The period to be used for the ADX calculation - default is 14"""
        if self.verb == 2:
            print(f"Calculating ADX with window: {window}")
        if df is None:
            df = self.df
        df['TR'] = np.max([
            df['High'] - df['Low'], 
            abs(df['High'] - df['Close'].shift(1)), 
            abs(df['Low'] - df['Close'].shift(1))
        ], axis=0)
        df['+DM'] = np.where((df['High'] - df['High'].shift(1)) > (df['Low'].shift(1) - df['Low']), df['High'] - df['High'].shift(1), 0)
        df['+DM'] = df['+DM'].where(df['+DM'] > 0, 0)
        df['-DM'] = np.where((df['Low'].shift(1) - df['Low']) > (df['High'] - df['High'].shift(1)), df['Low'].shift(1) - df['Low'], 0)
        df['-DM'] = df['-DM'].where(df['-DM'] > 0, 0)
        df['TR14'] = df['TR'].rolling(window=window).sum()
        df['+DM14'] = df['+DM'].rolling(window=window).sum()
        df['-DM14'] = df['-DM'].rolling(window=window).sum()
        df['+DI14'] = 100 * (df['+DM14'] / df['TR14'])
        df['-DI14'] = 100 * (df['-DM14'] / df['TR14'])
        df['DX'] = 100 * abs((df['+DI14'] - df['-DI14']) / (df['+DI14'] + df['-DI14']))
        df['ADX'] = df['DX'].rolling(window=window).mean()
        df.drop(['TR', '+DM', '-DM', 'TR14', '+DM14', '-DM14', '+DI14', '-DI14', 'DX'], axis=1, inplace=True)
        if self.verb > 0:
            print("ADX calculated")
        return df

    def calculate_VWAP(self, 
                       df:pd.DataFrame = None) -> pd.DataFrame:
        """Calculates the Volume Weighted Average Price of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs a 'Close', 'High', 'Low' and 'Volume' column - if None, the class dataframe is used
        """
        if self.verb == 2:
            print("Calculating VWAP")
        if df is None:
            df = self.df
        df['TP'] = (df['High'] + df['Low'] + df['Close']) / 3
        df['Traded'] = df['TP'] * df['Volume']
        df['Cumulative Traded'] = df['Traded'].cumsum()
        df['Cumulative Volume'] = df['Volume'].cumsum()
        df['VWAP'] = df['Cumulative Traded'] / df['Cumulative Volume']
        if self.verb > 0:
            print("VWAP calculated")
        return df        
    
    def calculate_standard_pivot_points(self, 
                                        df:pd.DataFrame = None) -> pd.DataFrame:
        """Calculates the Standard Pivot Points of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs 'High', 'Low' and 'Close' columns - if None, the class dataframe is used"""
        if self.verb == 2:
            print("Calculating Standard Pivot Points")
        if df is None:
            df = self.df
        df['Pivot'] = (df['High'] + df['Low'] + df['Close']) / 3
        df['R1'] = 2 * df['Pivot'] - df['Low']
        df['S1'] = 2 * df['Pivot'] - df['High']
        df['R2'] = df['Pivot'] + df['High'] - df['Low']
        df['S2'] = df['Pivot'] - df['High'] + df['Low']
        df['R3'] = df['Pivot'] + 2 * (df['High'] - df['Low'])
        df['S3'] = df['Pivot'] - 2 * (df['High'] - df['Low'])
        if self.verb > 0:
            print("Standard Pivot Points calculated")
        return df
    
    def calculate_fibonacci_pivot_points(self, 
                                         df:pd.DataFrame = None, 
                                         fib1:float = 0.382, 
                                         fib2:float = 0.618, 
                                         fib3: float = 1.382) -> pd.DataFrame:
        """Calculates the Fibonacci Pivot Points of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'High', 'Low' and 'Close' columns - if None, the class dataframe is used
        - fib1: float: The first fibonacci level - default is 0.382
        - fib2: float: The second fibonacci level - default is 0.618
        - fib3: float: The third fibonacci level - default is 1.382
        """
        if self.verb == 2:
            print(f"Calculating Fibonacci Pivot Points with levels: {fib1}, {fib2}, {fib3}")
        if df is None:
            df = self.df
        df['Pivot'] = (df['High'] + df['Low'] + df['Close']) / 3
        df['R1'] = df['Pivot'] + fib1 * (df['High'] - df['Low'])
        df['S1'] = df['Pivot'] - fib1 * (df['High'] - df['Low'])
        df['R2'] = df['Pivot'] + fib2 * (df['High'] - df['Low'])
        df['S2'] = df['Pivot'] - fib2 * (df['High'] - df['Low'])
        df['R3'] = df['Pivot'] + fib3 * (df['High'] - df['Low'])
        df['S3'] = df['Pivot'] - fib3 * (df['High'] - df['Low'])
        if self.verb > 0:
            print("Fibonacci Pivot Points calculated")
        return df
    
    def calculate_woodie_pivot_points(self, 
                                      df:pd.DataFrame = None) -> pd.DataFrame:
        """Calculates the Woodie Pivot Points of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'High', 'Low' and 'Close' columns - if None, the class dataframe is used
        """
        if self.verb == 2:
            print("Calculating Woodie Pivot Points")
        if df is None:
            df = self.df
        df['Pivot'] = (2 * df['Close'] + df['High'] + df['Low']) / 4        
        df['R1'] = 2 * df['Pivot'] - df['Low']
        df['R2'] = df['Pivot'] + (df['High'] - df['Low'])  
        df['R3'] = df['R1'] + (df['High'] - df['Low']) 
        df['S1'] = 2 * df['Pivot'] - df['High']
        df['S2'] = df['Pivot'] - (df['High'] - df['Low'])  
        df['S3'] = df['S1'] - (df['High'] - df['Low'])  
        if self.verb > 0:
            print("Woodie Pivot Points calculated")
        return df
    
    def calculate_ichimoku_cloud(self, 
                                 df:pd.DataFrame = None, 
                                 conversion_line_window:int = 9, 
                                 base_line_window:int = 26, 
                                 lagging_span_window:int = 52, 
                                 displacement:int = 26) -> pd.DataFrame:
        """Calculates the Ichimoku Cloud of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'High', 'Low' and 'Close' columns - if None, the class dataframe is used
        - conversion_line_window: int: The period to be used for the conversion line calculation - default is 9
        - base_line_window: int: The period to be used for the base line calculation - default is 26
        - lagging_span_window: int: The period to be used for the lagging span calculation - default is 52
        - displacement: int: The displacement to be used for the lagging span calculation - default is 26
        """
        if self.verb == 2:
            print(f"Calculating Ichimoku Cloud with conversion line window: {conversion_line_window}, base line window: {base_line_window}, lagging span window: {lagging_span_window} and displacement: {displacement}")
        if df is None:
            df = self.df
        period_high = df['High'].rolling(conversion_line_window).max()
        period_low = df['Low'].rolling(conversion_line_window).min()
        df['Tenkan-sen'] = (period_high + period_low) / 2
        period_high = df['High'].rolling(base_line_window).max()
        period_low = df['Low'].rolling(base_line_window).min()
        df['Kijun-sen'] = (period_high + period_low) / 2
        df['Senkou Span A'] = ((df['Tenkan-sen'] + df['Kijun-sen']) / 2).shift(displacement)
        period_high = df['High'].rolling(lagging_span_window).max()
        period_low = df['Low'].rolling(lagging_span_window).min()
        df['Senkou Span B'] = ((period_high + period_low) / 2).shift(displacement)
        df['Chikou Span'] = df['Close'].shift(-displacement)
        if self.verb > 0:
            print("Ichimoku Cloud calculated")
        return df
    
    def calculate_chaikin_money_flow(self, 
                                     df:pd.DataFrame = None, 
                                     window:int = 20) -> pd.DataFrame:
        """Calculates the Chaikin Money Flow of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'High', 'Low', 'Close' and 'Volume' columns - if None, the class dataframe is used
        - window: int: The period to be used for the Chaikin Money Flow calculation - default is 20
        """
        if self.verb == 2:
            print(f"Calculating Chaikin Money Flow with window: {window}")
        if df is None:
            df = self.df
        df['MF Multiplier'] = (2 * df['Close'] - df['Low'] - df['High']) / (df['High'] - df['Low'])
        df['MF Volume'] = df['MF Multiplier'] * df['Volume']
        df['CMF'] = df['MF Volume'].rolling(window).sum() / df['Volume'].rolling(window).sum()
        if self.verb > 0:
            print("Chaikin Money Flow calculated")
        return df
    
    def calculate_force_index(self, 
                              df:pd.DataFrame = None) -> pd.DataFrame:
        """Calculates the Force Index of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'Close' and 'Volume' columns - if None, the class dataframe is used
        """
        if self.verb == 2:
            print("Calculating Force Index")
        if df is None:
            df = self.df
        df['Force Index'] = df['Close'].diff(1) * df['Volume']
        if self.verb > 0:
            print("Force Index calculated")
        return df
    
    def calculate_zigzag(self, df: pd.DataFrame = None, threshold: float = 0.1) -> pd.DataFrame:
        """Calculates the ZigZag of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'Close', 'High', and 'Low' column - if None, the class dataframe is used
        - threshold: float: The threshold to be used for the ZigZag calculation - default is 0.1
        """
        if hasattr(self, 'verb') and self.verb == 2:
            print(f"Calculating ZigZag with threshold: {threshold}")
        if df is None:
            df = self.df
        df['ZigZag'] = None
        last_pivot_price = df.iloc[0]['Close']
        df.at[df.index[0], 'ZigZag'] = last_pivot_price
        last_pivot_index = 0
        trend = None
        
        for i in range(1, len(df)):
            high_change = (df.iloc[i]['High'] - last_pivot_price) / last_pivot_price
            low_change = (df.iloc[i]['Low'] - last_pivot_price) / last_pivot_price
            if trend is None or (trend and low_change <= -threshold) or (not trend and high_change >= threshold):
                if trend is None:
                    trend = high_change >= threshold
                elif trend and low_change <= -threshold:
                    trend = False
                elif not trend and high_change >= threshold:
                    trend = True
                pivot_price = df.iloc[i]['High'] if trend else df.iloc[i]['Low']
                df.loc[df.index[last_pivot_index]: df.index[i], 'ZigZag'] = last_pivot_price
                last_pivot_price = pivot_price
                last_pivot_index = i
        df.loc[df.index[last_pivot_index]:, 'ZigZag'] = last_pivot_price
        
        if hasattr(self, 'verb') and self.verb > 0:
            print("ZigZag calculated")
        return df
    
    def calculate_stochastic_momentum_index(self, 
                                            df:pd.DataFrame = None, 
                                            n:int = 14, 
                                            m:int = 3, 
                                            x:int = 9) -> pd.DataFrame:
        """Calculates the Stochastic Momentum Index of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'Close', 'High' and 'Low' columns - if None, the class dataframe is used
        - n: int: The period to be used for the first EMA calculation - default is 14
        - m: int: The period to be used for the second EMA calculation - default is 3
        - x: int: The period to be used for the Signal Line calculation - default is 9
        """
        if self.verb == 2:
            print(f"Calculating Stochastic Momentum Index with n: {n}, m: {m} and x: {x}")
        if df is None:
            df = self.df
        midpoint = (df['High'] + df['Low']) / 2
        D = df['Close'] - midpoint
        D_SMA = D.rolling(window=n).mean()
        HL = df['High'] - df['Low']
        HL_SMA = HL.rolling(window=n).mean()
        D_SMA2 = D_SMA.rolling(window=m).mean()
        HL_SMA2 = HL_SMA.rolling(window=m).mean()
        df['SMI'] = (D_SMA2 / (HL_SMA2 / 2)) * 100
        df['SMI Signal Line'] = df['SMI'].rolling(window=x).mean()
        if self.verb > 0:
            print("Stochastic Momentum Index calculated")
        return df

    def calculate_keltner_channel(self, 
                                  df:pd.DataFrame = None, 
                                  window:int = 20, 
                                  atr_window:int = 10,
                                  multiplier:int = 1) -> pd.DataFrame:
        """Calculates the Keltner Channel of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'Close', 'High' and 'Low' columns - if None, the class dataframe is used
        - window: int: The period to be used for the EMA calculation - default is 20
        - atr_window: int: The period to be used for the ATR calculation - default is 10
        - multiplier: int: The multiplier to be used for the Keltner Channel calculation - default is 1
        """
        if self.verb == 2:
            print(f"Calculating Keltner Channel with window: {window}, atr_window: {atr_window} and multiplier: {multiplier}")
        if df is None:
            df = self.df
        df['TR'] = np.max([
            df['High'] - df['Low'], 
            abs(df['High'] - df['Close'].shift(1)), 
            abs(df['Low'] - df['Close'].shift(1))
        ], axis=0)
        df['ATR'] = df['TR'].rolling(window=atr_window).mean()
        df['Upper'] = df['EMA'] + (multiplier * df['ATR'])
        df['Lower'] = df['EMA'] - (multiplier * df['ATR'])
        if self.verb > 0:
            print("Keltner Channel calculated")
        return df
    
    def calculate_elder_force_index(self, 
                                    df:pd.DataFrame = None, 
                                    window:int = 13) -> pd.DataFrame:
        """Calculates the Elder Force Index of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'Close' and 'Volume' columns - if None, the class dataframe is used
        - window: int: The period to be used for the Elder Force Index calculation - default is 13
        """
        if self.verb == 2:
            print(f"Calculating Elder Force Index with window: {window}")
        if df is None:
            df = self.df
        df['Force Index'] = df['Close'].diff(window) * df['Volume']
        df['EMA'] = df['Force Index'].ewm(span=window, adjust=False).mean()
        if self.verb > 0:
            print("Elder Force Index calculated")
        return df
    
    def calculate_sd(self, 
                     df:pd.DataFrame = None, 
                     window:int = 20) -> pd.DataFrame:
        """Calculates the Standard Deviation of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'Close' column - if None, the class dataframe is used
        - window: int: The period to be used for the Standard Deviation calculation - default is 20
        """
        if self.verb == 2:
            print(f"Calculating Standard Deviation with window: {window}")
        if df is None:
            df = self.df
        df['SD'] = df['Close'].rolling(window).std()
        if self.verb > 0:
            print("Standard Deviation calculated")
        return df
    
    def calculate_detrended_price_oscillator(self, 
                                             df:pd.DataFrame = None, 
                                             window:int = 20) -> pd.DataFrame:
        """Calculates the Detrended Price Oscillator of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'Close' column - if None, the class dataframe is used
        - window: int: The period to be used for the Detrended Price Oscillator calculation - default is 20
        """
        if self.verb == 2:
            print(f"Calculating Detrended Price Oscillator with window: {window}")
        if df is None:
            df = self.df
        df['DPO'] = df['Close'] - df['Close'].rolling(window).mean().shift(int(window / 2) + 1)
        if self.verb > 0:
            print("Detrended Price Oscillator calculated")
        return df
    
    def calculate_commodity_channel_index(self, 
                                          df:pd.DataFrame = None, 
                                          window:int = 20) -> pd.DataFrame:
        """Calculates the Commodity Channel Index of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'Close', 'High' and 'Low' columns - if None, the class dataframe is used
        - window: int: The period to be used for the Commodity Channel Index calculation - default is 20
        """
        if self.verb == 2:
            print(f"Calculating Commodity Channel Index with window: {window}")
        if df is None:
            df = self.df
        TP = (df['High'] + df['Low'] + df['Close']) / 3
        df['CCI'] = (TP - TP.rolling(window).mean()) / (0.015 * TP.rolling(window).std())
        if self.verb > 0:
            print("Commodity Channel Index calculated")
        return df
    
    def calculate_CMO(self, 
                      df:pd.DataFrame = None, 
                      window:int = 9) -> pd.DataFrame:
        """Calculates the Chande Momentum Oscillator of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'Close' column - if None, the class dataframe is used
        - window: int: The period to be used for the Chande Momentum Oscillator calculation - default is 9
        """
        if self.verb == 2:
            print(f"Calculating Chande Momentum Oscillator with window: {window}")
        if df is None:
            df = self.df
        up = df['Close'].diff().apply(lambda x: x if x > 0 else 0).rolling(window=window).sum()
        down = df['Close'].diff().apply(lambda x: -x if x < 0 else 0).rolling(window=window).sum()
        df['CMO'] = 100 * ((up - down) / (up + down))
        if self.verb > 0:
            print("Chande Momentum Oscillator calculated")
        return df

    
    def calculate_market_facilitation_index(self, 
                                            df:pd.DataFrame = None) -> pd.DataFrame:
        """Calculates the Market Facilitation Index of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'High' and 'Low' columns - if None, the class dataframe is used
        """
        if self.verb == 2:
            print("Calculating Market Facilitation Index")
        if df is None:
            df = self.df
        df['MFI'] = (df['High'] - df['Low']) / df['Volume']
        if self.verb > 0:
            print("Market Facilitation Index calculated")
        return df

    def calculate_volume_oscillator(self, 
                                    df: pd.DataFrame = None, 
                                    short_window:int = 12, 
                                    long_window:int = 26) -> pd.DataFrame:
        """Calculates the Volume Oscillator of the 'Volume' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'Volume' column - if None, the class dataframe is used
        - short_window: int: The period to be used for the short EMA calculation - default is 12
        - long_window: int: The period to be used for the long EMA calculation - default is 26
        """
        if self.verb == 2:
            print(f"Calculating Volume Oscillator with short window: {short_window} and long window: {long_window}")
        if df is None:
            df = self.df
        fast_vol_ema = df['Volume'].ewm(span=short_window, adjust=False).mean()
        slow_vol_ema = df['Volume'].ewm(span=long_window, adjust=False).mean()
        df['Volume Oscillator'] = ((fast_vol_ema - slow_vol_ema) / slow_vol_ema) * 100
        if self.verb > 0:
            print("Volume Oscillator calculated")
        return df

    
    def calculate_trix(self, 
                       df:pd.DataFrame = None, 
                       window:int = 15) -> pd.DataFrame:
        """Calculates the TRIX of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'Close' column - if None, the class dataframe is used
        - window: int: The period to be used for the TRIX calculation - default is 15
        """
        if self.verb == 2:
            print(f"Calculating TRIX with window: {window}")
        if df is None:
            df = self.df
        single_ema = df['Close'].ewm(span=window, adjust=False).mean()
        double_ema = single_ema.ewm(span=window, adjust=False).mean()
        triple_ema = double_ema.ewm(span=window, adjust=False).mean()
        df['TRIX'] = triple_ema.pct_change() * 100
        if self.verb > 0:
            print("TRIX calculated")
        return df
    


    def calculate_know_sure_thing(self, 
                                  df:pd.DataFrame = None, 
                                  roc_short:int = 10, 
                                  roc_long: int = 15, 
                                  signal_period: int = 9) -> pd.DataFrame:
        """Calculates the Know Sure Thing of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'Close' column - if None, the class dataframe is used
        - roc_short: int: The period to be used for the short Rate of Change calculation - default is 10
        - roc_long: int: The period to be used for the long Rate of Change calculation - default is 15
        - signal_period: int: The period to be used for the Signal Line calculation - default is 9
        """
        if self.verb == 2:
            print(f"Calculating Know Sure Thing with short ROC: {roc_short}, long ROC: {roc_long} and signal period: {signal_period}")
        if df is None:
            df = self.df
        roc1 = df['Close'].diff(roc_short).rolling(roc_short).mean()
        roc2 = df['Close'].diff(roc_long).rolling(roc_long).mean()
        roc3 = df['Close'].diff(roc_long * 2).rolling(roc_long * 2).mean()
        roc4 = df['Close'].diff(roc_long * 3).rolling(roc_long * 3).mean()
        df['KST'] = roc1 + 2 * roc2 + 3 * roc3 + 4 * roc4
        df['KST Signal'] = df['KST'].rolling(signal_period).mean()
        if self.verb > 0:
            print("Know Sure Thing calculated")
        return df
    
    def calculate_mass_index(self, 
                             df:pd.DataFrame = None, 
                             ema_period:int = 9,
                             sum_window:int = 25) -> pd.DataFrame:
        """Calculates the Mass Index of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'High' and 'Low' columns - if None, the class dataframe is used
        - ema_period: int: The period to be used for the EMA calculation - default is 9
        - sum_window: int: The period to be used for the Sum calculation - default is 25
        """
        if self.verb == 2:
            print(f"Calculating Mass Index with EMA period: {ema_period} and sum window: {sum_window}")
        if df is None:
            df = self.df
        high_low_range = df['High'] - df['Low']
        single_ema = high_low_range.ewm(span=ema_period, adjust=False).mean()
        double_ema = single_ema.ewm(span=ema_period, adjust=False).mean()
        ema_ratio = single_ema / double_ema
        df['Mass Index'] = ema_ratio.rolling(window=sum_window).sum()
        if self.verb > 0:
            print("Mass Index calculated")
        return df

    def calculate_coppock_curve(self,
                                df:pd.DataFrame = None,
                                short_window:int = 11,
                                long_window:int = 14,
                                signal_period:int = 10) -> pd.DataFrame:
        """Calculates the Coppock Curve of the 'Close' column of the dataframe
        Inputs:
        - df: pd.DataFrame: The dataframe to be used - needs to have 'Close' column - if None, the class dataframe is used
        - short_window: int: The period to be used for the short Rate of Change calculation - default is 11
        - long_window: int: The period to be used for the long Rate of Change calculation - default is 14
        - signal_period: int: The period to be used for the Signal Line calculation - default is 10
        """
        if self.verb == 2:
            print(f"Calculating Coppock Curve with short window: {short_window}, long window: {long_window} and signal period: {signal_period}")
        def calculate_wma(df, period):
            weights = np.arange(1, period + 1)
            if self.verb == 2:
                print(f"Calculating WMA")
            return df.rolling(window=period).apply(lambda x: np.dot(x, weights) / weights.sum(), raw=True)

        if df is None:
            df = self.df
        roc1 = df['Close'].pct_change(short_window) * 100
        roc2 = df['Close'].pct_change(long_window) * 100
        roc_sum = roc1 + roc2
        df['Coppock Curve'] = calculate_wma(roc_sum, signal_period)
        if self.verb > 0:
            print("Coppock Curve calculated")
        return df
    
    def calculate_all(self, df: pd.DataFrame = None):
        """
        Calculates all the technical indicators in the class.
        Note: The class dataframe is used if no dataframe is provided.
        Should an external dataframe be provided, it should have all the expected columns.
        """
        if df is None:
            df = self.df
        counter = 0
        methods_to_call = [method for method in dir(self) if method.startswith('calculate_') and method != 'calculate_all']
        for method in methods_to_call:
            counter += 1
            if self.verb == 2:
                print(f"Calculating method {counter} out of {len(methods_to_call)}")
            func = getattr(self, method)
            new_df = func()
            for col in new_df.columns:
                if col not in df.columns:
                    df[col] = new_df[col]
        return df

    def __repr__(self) -> str:
        if self.ticker is not None:
            return f"Technical Indicators for {self.yf_stock.ticker}"
        elif self.ticker is None and self.df is not None:
            return "Technical Indicators for external dataframe"
        else:
            return "Technical Indicators class"



import pandas as pd
from statsmodels.tsa.statespace.sarimax import SARIMAX
import numpy as np
import optuna
from sklearn.model_selection import train_test_split
from sklearn.metrics import roc_auc_score

class _SMA(object):
    def __init__(self, target, window = 5):
        self.target = target
        self.window = window
        if not isinstance(self.target, pd.Series):
            self.target = pd.Series(self.target)
        if len(self.target) < self.window:
            raise ValueError("The window size is larger than the target size")
        if self.window < 1:
            raise ValueError("The window size must be greater than 0")
        if self.target.isna().sum() > 0:
            raise ValueError("The target contains missing values, please fill them before using this class")
    
    def predict_val(self, num_steps = 1):
        values = self.target.rolling(window = self.window).mean().shift(-1)
        return np.mean(values[-num_steps:])

class _EMA(object):
    def __init__(self, target, window = 5):
        self.target = target
        self.window = window
        if not isinstance(self.target, pd.Series):
            self.target = pd.Series(self.target)
        if len(self.target) < self.window:
            raise ValueError("The window size is larger than the target size")
        if self.window < 1:
            raise ValueError("The window size must be greater than 0")
        if self.target.isna().sum() > 0:
            raise ValueError("The target contains missing values, please fill them before using this class")
    
    def predict_val(self, num_steps = 1):
        values = self.target.ewm(span = self.window).mean().shift(-1).fillna(0)
        return np.mean(values[-num_steps:])

class _SARIMAX_model(object):
    def __init__(self, target: pd.Series, exog: pd.Series = None, optimize: bool = False, 
                 optimization_rounds: int = 100, test_size: float = 0.2, random_state: int = 42, verbose: bool = False,
                 p: int = 1, d: int = 1, q: int = 1, P: int = 1, D: int = 1, Q: int = 1, s: int = 1, trend: str = 'n',
                 measure_error: bool = False, time_varying_regression: bool = False, mle_regression: bool = False,
                 simple_differencing: bool = False, enforce_stationarity: bool = False, enforce_invertibility: bool = False,
                 hamilton_representation: bool = False, concentrate_scale: bool = False, trend_offset: int = 1,
                 use_exact_diffuse: bool = False):
        """
        Utility class to fit SARIMAX models with exogenous variables, optimize the hyperparameters and predict values
        Inputs:
            - target: pd.Series, target variable to predict
            - exog: pd.Series, exogenous variable to use in the model
            - optimize: bool, whether to optimize the hyperparameters
            - optimization_rounds: int, number of rounds to optimize the hyperparameters
            - test_size: float, size of the test set
            - random_state: int, random state for reproducibility
            - p: int, AR order
            - d: int, differencing order
            - q: int, MA order
            - P: int, seasonal AR order
            - D: int, seasonal differencing order
            - Q: int, seasonal MA order
            - s: int, seasonal period
            - trend: str, trend parameter
            - measure_error: bool, whether to measure the error
            - time_varying_regression: bool, whether to use time varying regression
            - mle_regression: bool, whether to use MLE regression
            - simple_differencing: bool, whether to use simple differencing
            - enforce_stationarity: bool, whether to enforce stationarity
            - enforce_invertibility: bool, whether to enforce invertibility
            - hamilton_representation: bool, whether to use Hamilton representation
            - concentrate_scale: bool, whether to concentrate scale
            - trend_offset: int, trend offset
            - use_exact_diffuse: bool, whether to use exact diffuse
        """
        self.assert_types(target, exog, optimize, optimization_rounds, test_size, random_state, p, d, q, P, D, Q, s, trend,
                            measure_error, time_varying_regression, mle_regression, simple_differencing, enforce_stationarity,
                            enforce_invertibility, hamilton_representation, concentrate_scale, trend_offset, use_exact_diffuse)
        self.target = target
        self.verbose = verbose
        self.exog = exog
        self.model = None
        if exog is not None:
            self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(self.exog, self.target, test_size = test_size, random_state = random_state)
        else:
            _, _, self.y_train, self.y_test = train_test_split(np.zeros(len(self.target)), self.target, test_size = test_size, random_state = random_state)
        if optimize:
            self.best_params = self.optimize(optimization_rounds)
            self.init_model(self.best_params['p'], self.best_params['q'], self.best_params['d'], self.best_params['D'],
                            self.best_params['P'], self.best_params['Q'], self.best_params['s'], self.best_params['trend'],
                            self.best_params['measure_error'], self.best_params['time_varying_regression'], self.best_params['mle_regression'],
                            self.best_params['simple_differencing'], self.best_params['enforce_stationarity'], self.best_params['enforce_invertibility'],
                            self.best_params['hamilton_representation'], self.best_params['concentrate_scale'], self.best_params['trend_offset'],
                            self.best_params['use_exact_diffuse'])
        else:
            self.init_model(p, q, d, D, P, Q, s, trend, measure_error, time_varying_regression, mle_regression,
                            simple_differencing, enforce_stationarity, enforce_invertibility, hamilton_representation,
                            concentrate_scale, trend_offset, use_exact_diffuse)
    
    def assert_types(self, target: pd.Series, exog: pd.Series = None, optimize: bool = False, 
                 optimization_rounds: int = 100, test_size: float = 0.2, random_state: int = 42,
                 p: int = 1, d: int = 1, q: int = 1, P: int = 1, D: int = 1, Q: int = 1, s: int = 1, trend: str = 'n',
                 measure_error = False, time_varying_regression = False, mle_regression = False,
                 simple_differencing = False, enforce_stationarity = False, enforce_invertibility = False,
                 hamilton_representation = False, concentrate_scale = False, trend_offset = 1,
                 use_exact_diffuse = False):
        """
        Utility method to assert the types of the inputs
        """
        if not isinstance(target, pd.Series):
            raise ValueError("The target variable must be a pandas Series")
        if exog is not None:
            if not isinstance(exog, pd.Series):
                raise ValueError("The exogenous variable must be a pandas Series")
        if len(target) != len(exog):
            raise ValueError("The target and exogenous variables must have the same length")
        if not isinstance(optimize, bool):
            raise ValueError("The optimize parameter must be a boolean")
        if not isinstance(optimization_rounds, int):
            raise ValueError("The optimization_rounds parameter must be an integer")
        if not isinstance(test_size, float) or test_size > 1 or test_size < 0:
            raise ValueError("The test_size parameter must be a float between 0 and 1")
        if not isinstance(random_state, int):
            raise ValueError("The random_state parameter must be an integer")
        if not isinstance(p, int):
            raise ValueError("The p parameter must be an integer")
        if not isinstance(d, int):
            raise ValueError("The d parameter must be an integer")
        if not isinstance(q, int):
            raise ValueError("The q parameter must be an integer")
        if not isinstance(P, int):
            raise ValueError("The P parameter must be an integer")
        if not isinstance(D, int):
            raise ValueError("The D parameter must be an integer")
        if not isinstance(Q, int):
            raise ValueError("The Q parameter must be an integer")
        if not isinstance(s, int):
            raise ValueError("The s parameter must be an integer")
        if trend not in ['n', 'c', 't', 'ct']:
            raise ValueError("The trend parameter must be one of 'n', 'c', 't', 'ct'")
        if not isinstance(measure_error, bool):
            raise ValueError("The measure_error parameter must be a boolean")
        if not isinstance(time_varying_regression, bool):
            raise ValueError("The time_varying_regression parameter must be a boolean")
        if not isinstance(mle_regression, bool):
            raise ValueError("The mle_regression parameter must be a boolean")
        if not isinstance(simple_differencing, bool):
            raise ValueError("The simple_differencing parameter must be a boolean")
        if not isinstance(enforce_stationarity, bool):
            raise ValueError("The enforce_stationarity parameter must be a boolean")
        if not isinstance(enforce_invertibility, bool):
            raise ValueError("The enforce_invertibility parameter must be a boolean")
        if not isinstance(hamilton_representation, bool):
            raise ValueError("The hamilton_representation parameter must be a boolean")
        if not isinstance(concentrate_scale, bool):
            raise ValueError("The concentrate_scale parameter must be a boolean")
        if not isinstance(trend_offset, int):
            raise ValueError("The trend_offset parameter must be an integer")
        if not isinstance(use_exact_diffuse, bool):
            raise ValueError("The use_exact_diffuse parameter must be a boolean")
        
    def init_model(self, p, q, d, D, P, Q, s, 
                   trend, measurement_error, 
                   time_varying_regression, 
                   mle_regression, simple_differencing, 
                   enforce_stationarity, enforce_invertibility, 
                   hamilton_representation, concentrate_scale, 
                   trend_offset, use_exact_diffuse):
        """
        Method to initialize the SARIMAX model and fit it to the data
        """
        if self.verbose:
            print("Initializing model")
        if self.X_train is not None:
            self.model = SARIMAX(self.y_train, self.X_train, order = (p, d, q), seasonal_order = (P, D, Q, s), trend = trend, 
                                 measurement_error = measurement_error, time_varying_regression = time_varying_regression,
                                 mle_regression = mle_regression, simple_differencing = simple_differencing,
                                 enforce_stationarity = enforce_stationarity, enforce_invertibility = enforce_invertibility,
                                 hamilton_representation = hamilton_representation, concentrate_scale = concentrate_scale,
                                 trend_offset = trend_offset, use_exact_diffuse = use_exact_diffuse)
        else:
            self.model = SARIMAX(self.y_train, order = (p, d, q), seasonal_order = (P, D, Q, s), trend = trend, 
                                 measurement_error = measurement_error, time_varying_regression = time_varying_regression,
                                 mle_regression = mle_regression, simple_differencing = simple_differencing,
                                 enforce_stationarity = enforce_stationarity, enforce_invertibility = enforce_invertibility,
                                 hamilton_representation = hamilton_representation, concentrate_scale = concentrate_scale,
                                 trend_offset = trend_offset, use_exact_diffuse = use_exact_diffuse)
        self.model.fit()    
    
    def optimize(self, rounds):
        print("Warning: This feature is experimental, it may take a long time to run, as some models may be invalid or take a long time to fit")
        def objective(trial):
            p = trial.suggest_int('p', 0, 5)
            q = trial.suggest_int('q', 0, 5)
            d = trial.suggest_int('d', 0, 5)
            P = trial.suggest_int('P', 0, 5)
            Q = trial.suggest_int('Q', 0, 5)
            D = trial.suggest_int('D', 0, 5)
            s = trial.suggest_int('s', 0, 5)
            trend = trial.suggest_categorical('trend', ['n', 'c', 't', 'ct'])
            measure_error = trial.suggest_categorical('measure_error', [True, False])
            time_varying_regression = trial.suggest_categorical('time_varying_regression', [True, False])
            mle_regression = trial.suggest_categorical('mle_regression', [True, False])
            simple_differencing = trial.suggest_categorical('simple_differencing', [True, False])
            enforce_stationarity = trial.suggest_categorical('enforce_stationarity', [True, False])
            enforce_invertibility = trial.suggest_categorical('enforce_invertibility', [True, False])
            hamiltion_representation = trial.suggest_categorical('hamilton_representation', [True, False])
            concentrate_scale = trial.suggest_categorical('concentrate_scale', [True, False])
            trend_offset = trial.suggest_int('trend_offset', 0, 5)
            use_exact_diffuse = trial.suggest_categorical('use_exact_diffuse', [True, False])
            if self.verbose:
                print("Trying parameters: ", p, q, d, D, P, Q, s, trend, measure_error, time_varying_regression)
                print(mle_regression, simple_differencing, enforce_stationarity, enforce_invertibility, hamiltion_representation, concentrate_scale, trend_offset, use_exact_diffuse)
            try:
                self.init_model(p, q, d, D, P, Q, s, trend, measure_error, time_varying_regression, mle_regression,
                                simple_differencing, enforce_stationarity, enforce_invertibility, hamiltion_representation,
                                concentrate_scale, trend_offset, use_exact_diffuse)
                if self.X_test is not None:
                    prediction = self.model.forecast(steps = len(self.X_test), exog = self.X_test)
                else:
                    prediction = self.model.forecast(steps = len(self.y_test))
                if self.verbose:
                    print("The model was fitted successfully, the ROC AUC score is:", roc_auc_score(self.y_test, prediction))
                return roc_auc_score(self.y_test, prediction)
            except Exception as e:
                print("The model failed to fit, error: ", e)
                return 0
        study = optuna.create_study(direction = 'maximize')
        study.optimize(objective, n_trials = rounds)
        return study.best_params
    
    def forecast(self, steps: int = 1):
        """
        Method to forecast values
        Inputs:
            - steps: int, number of steps to forecast
        """
        return self.model.forecast(steps = steps)



"""Module containing test cases for the package."""
import unittest
from datafetcher import Fetcher as ftc 
import pandas as pd
from indicators import IndicatorCalculator

class TestDataFetcher(unittest.TestCase):
    """Module containing test cases for the package."""
    def test_historical_data_retrieval(self):
        """Test that historical data is retrieved without errors."""
        df = ftc("AAPL")
        self.assertIsNotNone(df.df, "Historical data should not be None")

    def test_invalid_ticker_handling(self):
        """Test how an invalid ticker is handled."""
        with self.assertRaises(AssertionError):
            ftc("INVALID_TICKER")
class TestIndicatorCalculator(unittest.TestCase):
    def setUp(self):
        self.sample_df = ftc('AAPL').df

    def test_initialization_with_ticker(self):
        calculator = IndicatorCalculator(ticker='AAPL')
        self.assertIsNotNone(calculator.df)

    def test_initialization_with_dataframe(self):
        calculator = IndicatorCalculator(dataframe=self.sample_df)
        self.assertEqual(len(calculator.df), 3)

    def test_calculate_rsi(self):
        calculator = IndicatorCalculator(dataframe=self.sample_df)
        calculator.calculate_RSI()
        self.assertIn('RSI', calculator.df.columns)

    def test_calculate_macd(self):
        calculator = IndicatorCalculator(dataframe=self.sample_df)
        calculator.calculate_MACD()
        self.assertIn('MACD', calculator.df.columns)
        self.assertIn('Signal', calculator.df.columns)

    def test_invalid_ticker(self):
        with self.assertRaises(ValueError):
            IndicatorCalculator(ticker='INVALID')

    def test_missing_columns(self):
        incomplete_df = pd.DataFrame({'Close': [100, 101, 102]})
        calculator = IndicatorCalculator(dataframe=incomplete_df)
        with self.assertRaises(KeyError):
            calculator.calculate_on_balance_volume()

    def test_verbosity(self):
        calculator = IndicatorCalculator(dataframe=self.sample_df, verbose=2)
        with self.assertLogs(level='INFO') as log:
            calculator.calculate_RSI()
            self.assertIn('Calculating RSI', log.output[0])

if __name__ == '__main__':
    unittest.main()



class MethodError(Exception):
    """
    Internal class to raise an error when an invalid method is passed, for internal use only (or not, but I don't see the point)
    """
    def __init__(self, message):
        self.message = message
        super().__init__(self.message)



"""
This module contains the Fetcher class which is designed to retrieve financial information. 
It uses the yfinance library to gather historical data for specified companies.
"""
from requests.exceptions import RequestException
import yfinance as yf
import pandas as pd
import requests
from typing import Tuple


class Fetcher:
    """
    A financial data retrieval utility designed to fetch and analyze financial 
    information for a specified company. This class offers methods to retrieve 
    historical stock data, financial statements (income statement, balance sheet, 
    and cash flow statement), and compute various financial metrics and ratios 
    critical for financial analysis.

    Attributes:
        ticker (str): The stock ticker symbol for the company.
        df (pd.DataFrame): DataFrame containing the company's historical stock data.
        yf_stock (yf.Ticker): yfinance Ticker object for accessing Yahoo Finance API data.
        income_statement (pd.DataFrame): Company's income statement.
        balance_sheet (pd.DataFrame): Company's balance sheet.
        cashflow (pd.DataFrame): Company's cash flow statement.

    Methods offer capabilities to:
    - Retrieve the company's ticker symbol based on its name.
    - Fetch historical price data and compute volatility.
    - Access detailed financial statements.
    - Calculate key financial metrics and ratios such as EPS, ROA, ROE, and more.
    - Evaluate the company's liquidity, leverage, and profitability.

    Usage:
    fetcher = Fetcher(ticker='AAPL')
    print(fetcher.get_pe_ratio())
    print(fetcher.get_current_ratio())

    Note:
    This class requires an internet connection to fetch data from Yahoo Finance.
    """
    def __init__(self, ticker:str = None, period:str = 'max'):
        if ticker:
            self.ticker = ticker
            self.df, self.yf_stock = self.get_historical_data(self.ticker, period)
            self.income_statement, self.balance_sheet, self.cashflow = self.get_financial_statements(self.yf_stock)
        else:
            print("Warning, no ticker has been specified. Only get_ticker(), get_risk_free_rate() and get_sp500() methods will work.")    
    
    def get_ticker(self, company_name:str)->str:
        """
        This function returns the ticker of a company, given the name.
        Input:
            -company_name: string of lenght > 4
        """
        yfinance = "https://query2.finance.yahoo.com/v1/finance/search"
        user_agent = (
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like"
            " Gecko) Chrome/108.0.0.0 Safari/537.36"
        )
        params = {"q": company_name, "quotes_count": 1, "country": "United States"}
        res = requests.get(
            url=yfinance, params=params, headers={"User-Agent": user_agent}, timeout = 50
        )
        data = res.json()
        if "quotes" in data and len(data["quotes"]) > 0:
            company_code = data["quotes"][0]["symbol"]
            return company_code
        raise AssertionError("Company may not be listed")
    
    def get_historical_data(self, symbol:str, timeframe:str)-> Tuple[pd.DataFrame, str]:
        """
        This function returns a PANDAS dataframe about the price of the company and the 
        stock symbol of the company input. The company must be listed in the NYSE.
        
        Input:
            -symbol: string with the name (or ticker) of the company in question; 
            capitalization is not regarded.
        """
        try:
            stock_symbol = symbol
            stock_symbol = self.get_ticker(stock_symbol)
            if not stock_symbol:
                return None
            stock = yf.Ticker(stock_symbol)
            historical_data = stock.history(period=timeframe)
            historical_data["Volatility"] = historical_data["Close"].rolling(window=30).std()
            historical_data = historical_data.dropna()
            return historical_data, stock
        except RequestException as e:     
            print(f"Network-related error occurred: {e}")
        return None, None
    
    def get_financial_statements(self, stock:yf.Ticker) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Given a yf.Ticker, it returns the three financial statements of the company, if present
        """
        try:
            income_statement = stock.financials
            balance_sheet = stock.balance_sheet
            cashflow = stock.cashflow
            return income_statement, balance_sheet, cashflow
        except RequestException as e:
            print(f"Network-related error occurred: {e}")
            return None, None, None
    
    def get_income_statement(self, symbol:str= None) -> pd.DataFrame:
        """Returns the income stament
        if "symbol" is not specified, it returns the income statement of the company the 
        Fetcher has been initialized to
        """
        if not symbol:
            return self.income_statement
        else:
            stock_symbol = symbol
            stock_symbol = self.get_ticker(stock_symbol)
            stock = yf.Ticker(stock_symbol)
            return stock.financials
    
    def get_balance_sheet(self, symbol:str= None) -> pd.DataFrame:
        """Returns the balance sheet
        if "symbol" is not specified, it returns the income statement of the company the 
        Fetcher has been initialized to
        """
        if not symbol:
            return self.balance_sheet
        else:
            stock_symbol = symbol
            stock_symbol = self.get_ticker(stock_symbol)
            stock = yf.Ticker(stock_symbol)
            return stock.balance_sheet
    
    def get_cashflow(self, symbol:str= None) -> pd.DataFrame:
        """Returns the cash flow
        if "symbol" is not specified, it returns the income statement of the company the 
        Fetcher has been initialized to
        """
        if not symbol:
            return self.cashflow
        else:
            stock_symbol = symbol
            stock_symbol = self.get_ticker(stock_symbol)
            stock = yf.Ticker(stock_symbol)
            return stock.cashflow
    
    def get_beta_value(self) -> float:
        """Returns the beta value of the stock"""
        return self.yf_stock.info['beta']    
    
    def get_pe_ratio(self) -> float:
        """Returns the price to earnings ratio"""
        return self.yf_stock.info['trailingPE']
    
    def get_ebitda(self) -> float:
        """Returns the EBITDA of the stock"""
        return self.income_statement.loc['EBITDA'].iloc[0]
    
    def get_ebit(self) -> float:
        """Returns the EBIT of the stock"""
        return self.income_statement.loc['EBIT'].iloc[0]

    def get_normalized_ebitda(self) -> float:
        """Returns the normalized EBITDA of the stock"""
        return self.income_statement.loc['Normalized EBITDA'].iloc[0]
    
    def get_gross_profit(self) -> float:
        """Returns the gross profit of the stock"""
        return self.income_statement.loc['Gross Profit'].iloc[0]
    
    def get_gross_profit_margin(self) -> float:
        """Returns the gross profit margin of the stock"""
        revenue = self.income_statement.loc['Total Revenue'].iloc[0]
        gross_profit = self.get_gross_profit()
        return (gross_profit / revenue) * 100
    
    def get_cogs_trend(self) -> float:
        """Returns the cost of goods sold"""
        cogs = self.income_statement.loc['Cost of Revenue']
        return cogs.pct_change()
    
    def get_roa(self) -> float:
        """Returns the return on assets"""
        net_income = self.income_statement.loc['Net Income'].iloc[0]
        total_assets = self.balance_sheet.loc['Total Assets'].iloc[0]
        return (net_income / total_assets) * 100
    
    def get_roe(self) -> float:
        """Returns the return on equity"""
        net_income = self.income_statement.loc['Net Income'].iloc[0]
        total_equity = self.balance_sheet.loc['Total Equity'].iloc[0]
        return (net_income / total_equity) * 100
    
    def get_total_current_assets(self) -> float:
        """Returns the current assets"""
        values_to_sum = []
        current_assets = 0
        for value in self.balance_sheet.index:
            if 'Current' in value and 'Assets' in value:
                values_to_sum.append(value)
        for value in values_to_sum:
            current_assets += self.balance_sheet.loc[value].iloc[0]
        return current_assets
        
    def get_total_current_liabilities(self) -> float:
        """Returns the current liabilities"""
        values_to_sum = []
        current_liabilities = 0
        for value in self.balance_sheet.index:
            if 'Current' in value and 'Liabilities' in value:
                values_to_sum.append(value)
        for value in values_to_sum:
            current_liabilities += self.balance_sheet.loc[value].iloc[0]
        return current_liabilities
    
    def get_current_ratio(self) -> float:
        """Returns the current ratio"""
        current_assets = self.get_total_current_assets()
        current_liabilities = self.get_total_current_liabilities()
        return (current_assets / current_liabilities) * 100

    def get_quick_ratio(self) -> float:
        """Returns the quick ratio"""
        current_assets = self.get_total_current_assets()
        inventory = self.balance_sheet.loc['Inventory'].iloc[0]
        current_liabilities = self.get_total_current_liabilities()
        return ((current_assets - inventory) / current_liabilities) * 100

    def get_debt_to_equity(self) -> float:
        """Returns the debt to equity ratio"""
        total_debt = self.balance_sheet.loc['Total Debt'].iloc[0]
        total_equity = self.balance_sheet.loc['Stockholders Equity'].iloc[0]
        return (total_debt / total_equity) * 100
    
    def get_interest_coverage(self) -> float:
        """Returns the interest coverage ratio"""
        ebit = self.get_ebit()
        interest_expense = self.income_statement.loc['Interest Expense'].iloc[0]
        return (ebit / interest_expense) * 100
    
    def get_cash_ratio(self) -> float:
        """Returns the cash ratio"""
        cash = self.balance_sheet.loc['Cash And Cash Equivalents'].iloc[0]
        current_liabilities = self.get_total_current_liabilities()
        return (cash / current_liabilities) * 100
    
    def get_net_working_capital(self) -> float:
        """Returns the net working capital"""
        current_assets = self.get_total_current_assets()
        current_liabilities = self.get_total_current_liabilities()
        return current_assets - current_liabilities
    
    def get_net_working_capital_ratio(self) -> float:
        """Returns the net working capital ratio"""
        net_working_capital = self.get_net_working_capital()
        current_liabilities = self.get_total_current_liabilities()
        return net_working_capital/ current_liabilities
    
    def get_eps(self) -> float:
        """Returns the earnings per share"""
        return self.income_statement.loc['Basic EPS'].iloc[0]
    
    def get_debt_ratio(self) -> float:
        """Returns the debt ratio"""
        total_debt = self.balance_sheet.loc['Total Debt'].iloc[0]
        total_assets = self.balance_sheet.loc['Total Assets'].iloc[0]
        return (total_debt / total_assets) * 100
    
    def get_equity_multiplier(self) -> float:
        """Returns the equity multiplier"""
        total_assets = self.balance_sheet.loc['Total Assets'].iloc[0]
        total_equity = self.balance_sheet.loc['Stockholders Equity'].iloc[0]
        return total_assets / total_equity
    
    def get_price_to_book(self) -> float:
        """Returns the price to book ratio"""
        book_value = self.balance_sheet.loc['Stockholders Equity'].iloc[0] / self.yf_stock.info['sharesOutstanding']
        return self.yf_stock.info['marketCap'] / book_value
    
    def get_price_to_sales(self) -> float:
        """Returns the price to sales ratio"""
        revenue = self.income_statement.loc['Total Revenue'].iloc[0]
        return self.yf_stock.info['marketCap'] / revenue

    def get_risk_free_rate(self, horizon:str = '13 weeks') -> float:
        """Returns the risk free rate, given an horizon.
        Accepted inputs are:
            -13 weeks
            -5 years
            -10 years
            -30 years
        """
        if horizon == '13 weeks':
            to_ret = yf.Ticker('^IRX').history(period='max')
            to_ret['Close'] = to_ret['Close']/100
            return to_ret
        elif horizon == '5 years':
            to_ret = yf.Ticker('^FVX').history(period='max')
            to_ret['Close'] = to_ret['Close']/100
            return to_ret
        elif horizon == '10 years':
            to_ret = yf.Ticker('^TNX').history(period='max')
            to_ret['Close'] = to_ret['Close']/100
            return to_ret
        elif horizon == '30 years':
            to_ret = yf.Ticker('^TYX').history(period='max')
            to_ret['Close'] = to_ret['Close']/100
            return to_ret
        else:
            raise ValueError('Invalid horizon, please choose between "month", "5 years", "10 years" and "30 years"')

    def get_sp500(self) -> pd.DataFrame:
        """Returns the S&P 500"""
        return yf.Ticker('^GSPC').history(period='max')
    
    def get_dividend_yield(self) -> float:
        """Returns the dividend yield"""
        return self.yf_stock.info['dividendYield']
    
    def get_payout_ratio(self) -> float:
        """Returns the payout ratio"""
        return self.yf_stock.info['payoutRatio']

    def get_earnings_growth_rate(self) -> float:
        """Returns the growth rate"""
        return self.yf_stock.info['earningsQuarterlyGrowth']
    
    def __repr__(self) -> str:
        try:
            return f"Stock: {self.yf_stock.ticker}\nLast Close Value: {self.df['Close'].iloc[-1]}\nBeta: {self.get_beta_value()}"
        except:
            return f"Stock: {self.yf_stock.ticker}\nLast Close Value: {self.df['Close'].iloc[-1]}\nBeta: Not available"
